""" opentrons.system.nmcli: Functions and data for interacting with nmcli

The functions contained here are for bridging Python calls with nmcli command
line invocations. They are in general not safe to call anywhere except an
Opentrons robot; on systems that do not have network-manager (like OSX, Windows
and some Linux distributions) they will not work, but on systems that _do_ they
may alter or destroy networking configurations.

In general, the functions here are light shims around nmcli invocations. This
is relevant in particular because they mostly do not handle exceptions coming
from subprocess itself, only parsing nmcli output.
"""

import csv
import logging
import re
import copy
from typing import Optional, List, Tuple, Dict, Callable, Any, NamedTuple
import enum
import os

from shlex import quote
from asyncio import subprocess as as_subprocess

from opentrons import config


log = logging.getLogger(__name__)


class EAPType(NamedTuple):
    name: str
    displayName: str
    args: List[Dict[str, Any]]


class _EAP_OUTER_TYPES(enum.Enum):
    """The types of phase-1 EAP we support."""

    # The string values of these supported EAP types should match both what
    # is expected by nmcli/wpa_supplicant (in 802-1x.eap) and the keys in
    # the CONFIG_REQUIRES dict above
    # Note: if a file field has an optional password, the password field `name`
    # must start with the `name` of the file field
    TLS = EAPType(
        name="tls",
        displayName="EAP-TLS",
        args=[
            {
                "name": "identity",
                "displayName": "Username",
                "nmName": "identity",
                "required": True,
                "type": "string",
            },
            {
                "name": "caCert",
                "displayName": "CA Certificate File",
                "nmName": "ca-cert",
                "required": False,
                "type": "file",
            },
            {
                "name": "clientCert",
                "displayName": "Client Certificate File",
                "nmName": "client-cert",
                "required": True,
                "type": "file",
            },
            {
                "name": "privateKey",
                "displayName": "Private Key File",
                "nmName": "private-key",
                "required": True,
                "type": "file",
            },
            {
                "name": "privateKeyPassword",
                "displayName": "Private Key Password",
                "nmName": "private-key-password",
                "required": False,
                "type": "password",
            },
        ],
    )
    PEAP = EAPType(
        name="peap",
        displayName="EAP-PEAP",
        args=[
            {
                "name": "identity",
                "displayName": "Username",
                "nmName": "identity",
                "required": True,
                "type": "string",
            },
            {
                "name": "anonymousIdentity",
                "displayName": "Anonymous Identity",
                "nmName": "anonymous-identity",
                "required": False,
                "type": "string",
            },
            {
                "name": "caCert",
                "displayName": "CA Certificate File",
                "nmName": "ca-cert",
                "required": False,
                "type": "file",
            },
        ],
    )
    TTLS = EAPType(
        name="ttls",
        displayName="EAP-TTLS",
        args=[
            {
                "name": "identity",
                "displayName": "Username",
                "nmName": "identity",
                "required": True,
                "type": "string",
            },
            {
                "name": "anonymousIdentity",
                "displayName": "Anonymous Identity",
                "nmName": "anonymous-identity",
                "required": False,
                "type": "string",
            },
            {
                "name": "caCert",
                "displayName": "CA Certificate File",
                "nmName": "ca-cert",
                "required": False,
                "type": "file",
            },
            {
                "name": "clientCert",
                "displayName": "Client Certificate File",
                "nmName": "client-cert",
                "required": False,
                "type": "file",
            },
            {
                "name": "privateKey",
                "displayName": "Private Key File",
                "nmName": "private-key",
                "required": False,
                "type": "file",
            },
            {
                "name": "privateKeyPassword",
                "displayName": "Private Key Password",
                "nmName": "private-key-password",
                "required": False,
                "type": "password",
            },
        ],
    )

    def qualified_name(self) -> str:
        return self.value.name


class _EAP_PHASE2_TYPES(enum.Enum):
    """The types of EAP phase 2 auth (for tunneled EAP) we support"""

    # The string values of these supported EAP types should match both what
    # is expected by nmcli/wpa_supplicant (in 802-1x.phase2-autheap) and the
    # keys in the CONFIG_REQUIRES dict above
    MSCHAP_V2 = EAPType(
        name="mschapv2",
        displayName="MS-CHAP v2",
        args=[
            {
                "name": "password",
                "displayName": "Password",
                "nmName": "password",
                "required": True,
                "type": "password",
            }
        ],
    )
    MD5 = EAPType(
        name="md5",
        displayName="MD5",
        args=[
            {
                "name": "password",
                "displayName": "Password",
                "nmName": "password",
                "required": True,
                "type": "password",
            }
        ],
    )
    TLS = EAPType(
        name="tls",
        displayName="TLS",
        args=[
            {
                "name": "phase2CaCert",
                "displayName": "Inner CA Certificate File",
                "nmName": "phase2-ca-cert",
                "required": False,
                "type": "file",
            },
            {
                "name": "phase2ClientCert",
                "displayName": "Inner Client Certificate File",
                "nmName": "phase2-client-cert",
                "required": True,
                "type": "file",
            },
            {
                "name": "phase2PrivateKey",
                "displayName": "Inner Private Key File",
                "nmName": "phase2-private-key",
                "required": True,
                "type": "file",
            },
            {
                "name": "phase2PrivateKeyPassword",
                "displayName": "Inner Private Key Password",
                "nmName": "phase2-private-key-password",
                "required": False,
                "type": "password",
            },
        ],
    )

    def qualified_name(self) -> str:
        return "eap-" + self.value.name


class EAP_TYPES(enum.Enum):
    """The types of EAP we support, fusing inner and outer methods"""

    TTLS_EAPTLS = (_EAP_OUTER_TYPES.TTLS, _EAP_PHASE2_TYPES.TLS)
    TTLS_EAPMSCHAPV2 = (_EAP_OUTER_TYPES.TTLS, _EAP_PHASE2_TYPES.MSCHAP_V2)
    TTLS_MD5 = (_EAP_OUTER_TYPES.TTLS, _EAP_PHASE2_TYPES.MD5)
    PEAP_EAPMSCHAPV2 = (_EAP_OUTER_TYPES.PEAP, _EAP_PHASE2_TYPES.MSCHAP_V2)
    TLS = (_EAP_OUTER_TYPES.TLS, None)

    def __init__(self, outer: _EAP_OUTER_TYPES, inner: _EAP_PHASE2_TYPES) -> None:
        self.outer = outer
        self.inner = inner

    def qualified_name(self) -> str:
        name = self.outer.qualified_name()
        if self.inner:
            name += "/" + self.inner.qualified_name()
        return name

    @classmethod
    def by_qualified_name(cls, qname: str) -> "EAP_TYPES":
        for val in cls.__members__.values():
            if val.qualified_name() == qname:
                return val
        raise KeyError(qname)

    def args(self) -> List[Dict[str, Any]]:
        # Have to copy these or reference semantics modify the version stored
        # in the enums
        to_ret = copy.deepcopy(self.outer.value.args)
        if self.inner:
            to_ret += copy.deepcopy(self.inner.value.args)
        return to_ret

    def display_name(self) -> str:
        name = self.outer.value.displayName
        if self.inner:
            name += " with " + self.inner.value.displayName
        return name


class SECURITY_TYPES(enum.Enum):
    """The types of security that this module supports."""

    # The string values of these supported security types are passed
    # directly to nmcli; they should match the security types allowed
    # in the network-manager settings for 802-11-wireless-security.key-mgmt
    NONE = "none"
    WPA_PSK = "wpa-psk"
    WPA_EAP = "wpa-eap"


class CONNECTION_TYPES(enum.Enum):
    """Types of connection (used to parse nmcli results)"""

    # These connection types are used to parse nmcli results and should be
    # valid results for the last element when splitting connection.type by ’-’
    WIRELESS = "wireless"
    ETHERNET = "ethernet"


class NETWORK_IFACES(enum.Enum):
    """Network interface names that we manage here."""

    WIFI = {
        config.SystemArchitecture.BUILDROOT: "wlan0",
        config.SystemArchitecture.YOCTO: "mlan0",
    }.get(config.ARCHITECTURE, "wlan0")
    ETH_LL = "eth0"


def _add_security_type_to_scan(scan_out: Dict[str, Any]) -> Dict[str, Any]:
    sec = scan_out["security"]
    if "802.1X" in sec:
        scan_out["securityType"] = "wpa-eap"
    elif "WPA2" in sec:
        scan_out["securityType"] = "wpa-psk"
    elif "" == sec:
        scan_out["securityType"] = "none"
    else:
        scan_out["securityType"] = "unsupported"
    return scan_out


async def available_ssids(rescan: Optional[bool] = False) -> List[Dict[str, Any]]:
    """List the visible (broadcasting SSID) wireless networks.

    rescan: Forces a rescan instead of using the cached results.

    Returns a list of the SSIDs. They may contain spaces and should be escaped
    if later passed to a shell.
    """
    # Force nmcli to actually scan if rescan is true rather than reuse cached results.
    # We ignore errors here because NetworkManager yells at you if you do it twice in a
    # row without another operation in between
    if rescan:
        cmd = ["device", "wifi", "rescan"]
        _1, _2 = await _call(cmd, suppress_err=True)

    fields = ["ssid", "signal", "active", "security"]
    cmd = ["--terse", "--fields", ",".join(fields), "device", "wifi", "list"]
    out, err = await _call(cmd)
    if err:
        raise RuntimeError(err)
    output = _dict_from_terse_tabular(
        fields,
        out,
        transformers={
            "signal": lambda s: int(s) if s.isdigit() else None,
            "active": lambda a: a.lower() == "yes",
            "ssid": lambda s: s if s != "--" else None,
        },
    )

    return [_add_security_type_to_scan(nw) for nw in output if nw["ssid"]]


async def is_connected() -> str:
    """Return nmcli's connection measure: none/portal/limited/full/unknown"""
    res, _ = await _call(["networking", "connectivity", "check"])
    return res


async def connections(
    for_type: Optional[CONNECTION_TYPES] = None,
) -> List[Dict[str, str]]:
    """Return the list of configured connections.

    This is all connections that nmcli knows about and manages.
    Each connection is a dict containing some basic information - the
    information retrievable from nmcli connection show. Further information
    should be queried on a connection by connection basis.

    If for_type is not None, it should be a str containing an element of
    CONNECTION_TYPES, and results will be limited to that connection type.
    """
    fields = ["name", "type", "active"]
    res, _ = await _call(["-t", "-f", ",".join(fields), "connection", "show"])
    found = _dict_from_terse_tabular(
        fields,
        res,
        # ’ethernet’ or ’wireless’ from ’802-11-wireless’ or ’802-4-ethernet’
        # and bools from ’yes’ or ’no
        transformers={
            "type": lambda s: s.split("-")[-1],
            "active": lambda s: s.lower() == "yes",
        },
    )
    if for_type is not None:
        should_return = []
        for c in found:
            if c["type"] == for_type.value:
                should_return.append(c)
        return should_return
    else:
        return found


async def connection_exists(ssid: str) -> Optional[str]:
    """If there is already a connection for this ssid, return the name of
    the connection; if there is not, return None.
    """
    nmcli_conns = await connections()
    for wifi in [c["name"] for c in nmcli_conns if c["type"] == "wireless"]:
        res, _ = await _call(
            [
                "-t",
                "-f",
                "802-11-wireless.ssid",
                "-m",
                "tabular",
                "connection",
                "show",
                wifi,
            ]
        )
        if res == ssid:
            return wifi
    return None


async def _trim_old_connections(
    new_name: str, con_type: CONNECTION_TYPES
) -> Tuple[bool, str]:
    """Delete all connections of con_type but the one specified."""
    existing_cons = await connections(for_type=con_type)
    not_us = [c["name"] for c in existing_cons if c["name"] != new_name]
    ok = True
    res = []
    for c in not_us:
        this_ok, remove_res = await remove(name=c)
        ok = ok and this_ok
        if not this_ok:
            # This is not a failure case for connection, and indeed the new
            # connection is already established, so just warn about it
            log.warning("Could not remove wifi connection {}: {}".format(c, remove_res))
            res.append(remove_res)
        else:
            log.debug("Removed old wifi connection {}".format(c))
    return ok, ";".join(res)


def _add_eap_args(eap_args: Dict[str, str]) -> List[str]:
    """Add configuration options suitable for an nmcli con add command
    for WPA-EAP configuration. These options are mostly in the
    802-1x group.

    The eap_args dict should be a flat structure of arguments. They
    must contain at least 'eapType', specifying the EAP type to use
    (the qualified_name() of one of the members of EAP_TYPES) and the
    required arguments for that EAP type.
    """
    args = ["wifi-sec.key-mgmt", "wpa-eap"]
    eap_type = EAP_TYPES.by_qualified_name(eap_args["eapType"])
    type_args = eap_type.args()
    args += ["802-1x.eap", eap_type.outer.value.name]
    if eap_type.inner:
        args += ["802-1x.phase2-autheap", eap_type.inner.value.name]
    for ta in type_args:
        if ta["name"] in eap_args:
            if ta["type"] == "file":
                # Keyfiles must be prepended with file:// so nm-cli
                # knows that we’re not giving it DER-encoded blobs
                val = "file://" + eap_args[ta["name"]]
            else:
                val = eap_args[ta["name"]]
            args += ["802-1x." + ta["nmName"], val]
    return args


def _build_con_add_cmd(
    ssid: str,
    security_type: SECURITY_TYPES,
    psk: Optional[str],
    hidden: bool,
    eap_args: Optional[Dict[str, Any]],
) -> List[str]:
    """Build the nmcli connection add command to configure the new network.

    The parameters are the same as configure but without the defaults; this
    should be called only by configure.
    """
    configure_cmd = [
        "connection",
        "add",
        "save",
        "yes",
        "autoconnect",
        "yes",
        "ifname",
        NETWORK_IFACES.WIFI.value,
        "type",
        "wifi",
        "con-name",
        ssid,
        "wifi.ssid",
        ssid,
        "802-11-wireless.cloned-mac-address",
        "permanent",
    ]
    if hidden:
        configure_cmd += ["wifi.hidden", "true"]
    if security_type == SECURITY_TYPES.WPA_PSK:
        configure_cmd += ["wifi-sec.key-mgmt", security_type.value]
        if psk is None:
            raise ValueError("wpa-psk security type requires psk")
        configure_cmd += ["wifi-sec.psk", psk]
    elif security_type == SECURITY_TYPES.WPA_EAP:
        if eap_args is None:
            raise ValueError("wpa-eap security type requires eap_args")
        configure_cmd += _add_eap_args(eap_args)
    elif security_type == SECURITY_TYPES.NONE:
        pass
    else:
        raise ValueError("Bad security_type {}".format(security_type))

    return configure_cmd


async def configure(
    ssid: str,
    securityType: SECURITY_TYPES,
    psk: Optional[str] = None,
    hidden: bool = False,
    eapConfig: Optional[Dict[str, Any]] = None,
    upRetries: int = 2,
) -> Tuple[bool, str]:
    """Configure a connection but do not bring it up (though it is configured
    for autoconnect).

    Returns (success, message) where ``success`` is a ``bool`` and ``message``
    is a ``str``.

    Only anticipated failures are treated that way - for instance, an ssid
    that doesn't exist will get a False and a message; a system where nmcli
    is not found will raise a CalledProcessError.

    Input checks should be conducted before calling this function; any issues
    with arguments that make it to this function will probably surface
    themselves as TypeErrors and ValueErrors rather than anything more
    structured.

    The ssid and security_type arguments are mandatory; the others have
    different requirements depending on the security type.
    """
    already = await connection_exists(ssid)
    if already:
        # TODO(seth, 8/29/2018): We may need to do connection modifies
        # here for EAP configuration if e.g. we’re passing a keyfile in a
        # different http request
        _1, _2 = await _call(["connection", "delete", already])

    configure_cmd = _build_con_add_cmd(ssid, securityType, psk, hidden, eapConfig)
    res, err = await _call(configure_cmd)

    # nmcli connection add returns a string that looks like
    # "Connection ’connection-name’ (connection-uuid) successfully added."
    # This unfortunately doesn’t respect the --terse flag, so we need to
    # regex out the name or the uuid to use later in connection up; the
    # uuid is slightly more regular, so that’s what we use.
    uuid_matches = re.search(r"Connection '(.*)'[\s]+\(([\w\d-]+)\) successfully", res)
    if not uuid_matches:
        return False, err.split("\r")[-1]
    name = uuid_matches.group(1)
    uuid = uuid_matches.group(2)
    for _ in range(upRetries):
        res, err = await _call(["connection", "up", "uuid", uuid])
        if "Connection successfully activated" in res:
            # If we successfully added the connection, remove other wifi
            # connections so they don’t accumulate over time

            _3, _4 = await _trim_old_connections(name, CONNECTION_TYPES.WIRELESS)
            return True, res
    else:
        return False, err.split("\r")[-1]


async def wifi_disconnect(ssid: str) -> Tuple[bool, str]:
    """
    Disconnect from specified wireless network and delete the connection.

    Ideally, user would be allowed to disconnect a robot from wifi only over an
    ethernet connection to the robot so that, 1) they get the disconnect
    response back and 2) their robot isn't left with no connectivity at all
    However, with robot discovery (over eth0) issues still pending, we will
    allow the users to disconnect the robot from wifi regardless of whether
    they are connected via ethernet.

    Returns (True, msg) if the network was disconnected from successfully,
            (False, msg) otherwise
    """
    res, err = await _call(["connection", "down", ssid])
    if "successfully deactivated" in res:
        rem_ok, rem_res = await remove(ssid)
        if rem_ok:
            res = res + ".\n " + rem_res
        else:
            res = res + ".\n Error: Could not remove ssid. " + rem_res
        return True, res
    else:
        return False, err


async def remove(
    ssid: Optional[str] = None,
    name: Optional[str] = None,
) -> Tuple[bool, str]:
    """Remove a network. Depending on what is known, specify either ssid
    (in which case this function will call ``connection_exists`` to get the
    nmcli connection name) or the nmcli connection name directly.

    Returns (True, msg) if the connection was deleted, (False, msg) otherwise.
    """
    if None is not ssid:
        name = await connection_exists(ssid)
    if None is not name:
        res, err = await _call(["connection", "delete", name])
        if "successfully deleted" in res:
            return True, res
        else:
            return False, err
    else:
        return False, "No connection for ssid {}".format(ssid)


async def iface_info(which_iface: NETWORK_IFACES) -> Dict[str, Optional[str]]:
    """Get the basic network configuration of an interface.

    Returns a dict containing the info:
    {
      'ipAddress': 'xx.xx.xx.xx/yy' (ip4 addr with subnet as CIDR) or None
      'macAddress': 'aa:bb:cc:dd:ee:ff' or None
      'gatewayAddress: 'zz.zz.zz.zz' or None
    }

    which_iface should be a string in IFACE_NAMES.
    """
    # example device info lines
    #  GENERAL.HWADDR:B8:27:EB:24:D1:D0
    #  IP4.ADDRESS[1]:10.10.2.221/22
    # capture the field name (without the number in brackets) and the value
    # using regex instead of split because there may be ":" in the value
    _DEV_INFO_LINE_RE = re.compile(r"([\w.]+)(?:\[\d+])?:(.*)")
    # example device info: 30 (disconnected)
    # capture the string without the number
    _IFACE_STATE_RE = re.compile(r"\d+ \((.+)\)")

    info: Dict[str, Optional[str]] = {
        "ipAddress": None,
        "macAddress": None,
        "gatewayAddress": None,
        "state": None,
        "type": None,
    }
    fields = [
        "GENERAL.HWADDR",
        "IP4.ADDRESS",
        "IP4.GATEWAY",
        "GENERAL.TYPE",
        "GENERAL.STATE",
    ]
    # Note on this specific command: Most nmcli commands default to a tabular
    # output mode, where if there are multiple things to pull a couple specific
    # fields from it you’ll get a table where rows are, say, connections, and
    # columns are field name. However, specifically ‘con show <con-name>‘ and
    # ‘dev show <dev-name>’ default to a multiline representation, and even if
    # explicitly ask for it to be tabular, it’s not quite the same as the other
    # commands. So we have to special-case the parsing.
    res, err = await _call(
        [
            "--mode",
            "multiline",
            "--escape",
            "no",
            "--terse",
            "--fields",
            ",".join(fields),
            "dev",
            "show",
            which_iface.value,
        ]
    )

    field_map = {}
    for line in res.split("\n"):
        # pull the key (without brackets) and the value out of the line
        match = _DEV_INFO_LINE_RE.fullmatch(line)
        if match is None:
            raise ValueError("Bad nmcli result; out: {}; err: {}".format(res, err))
        key, val = match.groups()
        # nmcli can put "--" instead of "" for None
        field_map[key] = None if val == "--" else val

    info["macAddress"] = field_map.get("GENERAL.HWADDR")
    info["ipAddress"] = field_map.get("IP4.ADDRESS")
    info["gatewayAddress"] = field_map.get("IP4.GATEWAY")
    info["type"] = field_map.get("GENERAL.TYPE")
    state_val = field_map.get("GENERAL.STATE")

    if state_val:
        state_match = _IFACE_STATE_RE.fullmatch(state_val)
        if state_match:
            info["state"] = state_match.group(1)

    return info


async def _call(cmd: List[str], suppress_err: bool = False) -> Tuple[str, str]:
    """
    Runs the command in a subprocess and returns the captured stdout output.
    :param cmd: a list of arguments to nmcli. Should not include nmcli itself.

    :return: (stdout, stderr)
    """
    to_exec = [quote(c) for c in ["nmcli"] + cmd]
    cmd_str = " ".join(to_exec)
    # We have to use a shell invocation here because nmcli will not accept
    # secrets specified on the command line unless it’s in a shell. The other
    # option is editing the connection configuration file in /etc/ afterwards
    # (or using d-bus and pretending to be an auth agent)
    proc = await as_subprocess.create_subprocess_shell(
        cmd_str, stdout=as_subprocess.PIPE, stderr=as_subprocess.PIPE
    )
    out, err = await proc.communicate()
    out_str, err_str = out.decode().strip(), err.decode().strip()
    sanitized = sanitize_args(to_exec)
    log.debug("{}: stdout={}".format(" ".join(sanitized), out_str))
    if err_str and not suppress_err:
        log.warning("{}: stderr={}".format(" ".join(sanitized), err_str))
    return out_str, err_str


def sanitize_args(cmd: List[str]) -> List[str]:
    """Filter the command so that it no longer contains passwords"""
    sanitized = []
    for idx, fieldname in enumerate(cmd):

        def _is_password(cmdstr: str) -> bool:
            return "wifi-sec.psk" in cmdstr or "password" in cmdstr.lower()

        if idx > 0 and _is_password(cmd[idx - 1]):
            sanitized.append("****")
        else:
            sanitized.append(fieldname)
    return sanitized


def _parse_colonsep_response(response_fullstring: str) -> List[List[str]]:
    reader = csv.reader(response_fullstring.split("\n"), delimiter=":", escapechar="\\")
    return [row for row in reader if row]


def _dict_from_terse_tabular(
    names: List[str], inp: str, transformers: Dict[str, Callable[[str], Any]] = {}
) -> List[Dict[str, Any]]:
    """Parse NMCLI terse tabular output into a list of Python dict.

    ``names`` is a list of strings of field names to apply to the input data,
    which is assumed to be colon separated.

    ``inp`` is the input as a string (i.e. already decode()d) from nmcli

    ``transformers`` is a dict mapping field names to callables of the form
    f: str -> any. If a fieldname is in transformers, that callable will be
    invoked on the field matching the name and the result stored.

    The return value is a list with one element per valid line of input, where
    each element is a dict with keys taken from names and values from the input
    """
    res = []
    for n in names:
        if n not in transformers:
            transformers[n] = lambda s: s
    for fields in _parse_colonsep_response(inp):
        if len(fields) != len(names):
            log.warning(f"ignoring unparsable fields <{fields}>")
            continue
        res.append(
            dict(
                [
                    (elem[0], transformers[elem[0]](elem[1]))
                    for elem in zip(names, fields)
                ]
            )
        )
    return res


# The functions that follow are designed to configure the system so that the
# paths to the keys specified in `nmcli con add`, from our perspective in
# the container, are the same as the paths to the keys from the perspective
# of the host.

# When we run nmcli, it is talking via D-bus to the NetworkManager daemon
# running in the host - not in our container. That means that all the config
# lives in the host’s /etc/NetworkManager/system-connections (you can tell
# because when we have connections up, the containers
# /etc/NetworkManager/system-connections is empty - I think it only exists
# because nmcli creates it on install). That means that file paths must be
# correct from that perspective.

# In addition, `nmcli con add` checks locally, _in the container_, that the
# paths specified as key files do indeed contain keyfiles. That means
# whatever paths we specify must be valid in the container at the time that
# we call `nmcli con add`, and valid in the host when we call `nmcli con up`
# (i.e., all the time, including when the container isn’t running). So, we
# make the paths be what the host will see, and we create a symlink in the
# container to make the path look the same here.

# This path depends on the resin app ID, which cannot be hardcoded because
# it changes based on which Resin application the robot running the code
# is in. It is also specified by Resin, which means it is given to pid 1
# and relies on pid 1 passing it to children made with execve. To make this
# path specification work whenever we run this, including if somebody is
# running it directly from a shell which _won’t_ correctly pass the data,
# we need to parse the environment of pid1.


def _get_host_data_prefix() -> str:
    return os.path.join("mnt", "data", "resin-data", _get_resin_app_id())


def _get_resin_app_id() -> str:
    if not config.IS_ROBOT:
        raise RuntimeError("Resin app id is only available on the pi")
    p1_env = open("/proc/1/environ").read()
    # /proc/x/environ is pretty much just the raw memory segment of the
    # process that represents its environment. It contains a
    # NUL-separated list of strings.
    app_id = re.search("RESIN_APP_ID=([^\x00]*)\x00", p1_env)
    if not app_id:
        raise RuntimeError("Cannot find resin app id! /proc/1/env={}".format(p1_env))
    return app_id.group(1)


def _rewrite_key_path_to_host_path(key_path: str) -> str:
    resin_id = _get_resin_app_id()
    key_abspath = os.path.abspath(key_path)
    if key_abspath.startswith("/data"):
        key_relpath = os.path.relpath(key_abspath, "/data")
        key_hostpath = os.path.join(
            "/", "mnt", "data", "resin-data", resin_id, key_relpath
        )
        log.debug("Rewrote key path {} to {} for host".format(key_path, key_hostpath))
        return key_hostpath
    else:
        log.warning(
            "Wifi keys that are not in /data may not work unless"
            "they are specified in a path common to the host and"
            " the container"
        )
        return key_path


def _make_host_symlink_if_necessary() -> None:
    host_data_prefix = _get_host_data_prefix()
    if not os.path.islink(_get_host_data_prefix()):
        parent = os.path.abspath(os.path.join(host_data_prefix, os.pardir))
        os.makedirs(parent, exist_ok=True)
        os.symlink("/data", host_data_prefix)



"""Opentrons analyze CLI."""
import click

from anyio import run, Path as AsyncPath
from datetime import datetime, timezone
from pathlib import Path
from pydantic import BaseModel
from typing import Any, Dict, List, Optional, Sequence, Union
from typing_extensions import Literal

from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocol_reader import (
    ProtocolReader,
    ProtocolFileRole,
    ProtocolType,
    JsonProtocolConfig,
    ProtocolFilesInvalidError,
)
from opentrons.protocol_runner import create_simulating_runner
from opentrons.protocol_engine import (
    Command,
    ErrorOccurrence,
    LoadedLabware,
    LoadedPipette,
    LoadedModule,
    Liquid,
)

from opentrons_shared_data.robot.dev_types import RobotType


@click.command()
@click.argument(
    "files",
    nargs=-1,
    required=True,
    type=click.Path(exists=True, path_type=Path, file_okay=True, dir_okay=True),
)
@click.option(
    "--json-output",
    help="Return analysis results as machine-readable JSON.",
    type=click.Path(path_type=AsyncPath),
)
def analyze(files: Sequence[Path], json_output: Optional[Path]) -> None:
    """Analyze a protocol.

    You can use `opentrons analyze` to get a protocol's expected
    equipment and commands.
    """
    run(_analyze, files, json_output)


def _get_input_files(files_and_dirs: Sequence[Path]) -> List[Path]:
    results: List[Path] = []

    for entry in files_and_dirs:
        if entry.is_dir():
            results.extend(entry.glob("**/*"))
        else:
            results.append(entry)

    return results


async def _analyze(
    files_and_dirs: Sequence[Path],
    json_output: Optional[AsyncPath],
) -> None:
    input_files = _get_input_files(files_and_dirs)

    try:
        protocol_source = await ProtocolReader().read_saved(
            files=input_files,
            directory=None,
        )
    except ProtocolFilesInvalidError as error:
        raise click.ClickException(str(error))

    runner = await create_simulating_runner(
        robot_type=protocol_source.robot_type, protocol_config=protocol_source.config
    )
    analysis = await runner.run(deck_configuration=[], protocol_source=protocol_source)

    if json_output:
        results = AnalyzeResults.construct(
            createdAt=datetime.now(tz=timezone.utc),
            files=[
                ProtocolFile.construct(name=f.path.name, role=f.role)
                for f in protocol_source.files
            ],
            config=(
                JsonConfig.construct(
                    schemaVersion=protocol_source.config.schema_version
                )
                if isinstance(protocol_source.config, JsonProtocolConfig)
                else PythonConfig.construct(
                    apiVersion=protocol_source.config.api_version
                )
            ),
            metadata=protocol_source.metadata,
            robotType=protocol_source.robot_type,
            commands=analysis.commands,
            errors=analysis.state_summary.errors,
            labware=analysis.state_summary.labware,
            pipettes=analysis.state_summary.pipettes,
            modules=analysis.state_summary.modules,
            liquids=analysis.state_summary.liquids,
        )

        await json_output.write_text(
            results.json(exclude_none=True),
            encoding="utf-8",
        )

    else:
        raise click.UsageError(
            "Currently, this tool only supports JSON mode. Use `--json-output`."
        )


class ProtocolFile(BaseModel):
    """A file in a protocol analysis."""

    name: str
    role: ProtocolFileRole


class JsonConfig(BaseModel):
    """Configuration of a JSON protocol."""

    protocolType: Literal[ProtocolType.JSON] = ProtocolType.JSON
    schemaVersion: int


class PythonConfig(BaseModel):
    """Configuration of a Python protocol."""

    protocolType: Literal[ProtocolType.PYTHON] = ProtocolType.PYTHON
    apiVersion: APIVersion


class AnalyzeResults(BaseModel):
    """Results of a protocol analysis.

    See robot-server's analysis models for field documentation.
    """

    # We want to unify this local analysis model with the one that robot-server returns.
    # Until that happens, we need to keep these fields in sync manually.

    # Fields that are currently unique to this local analysis module, missing from robot-server:
    createdAt: datetime
    files: List[ProtocolFile]
    config: Union[JsonConfig, PythonConfig]
    metadata: Dict[str, Any]

    # Fields that should match robot-server:
    robotType: RobotType
    commands: List[Command]
    labware: List[LoadedLabware]
    pipettes: List[LoadedPipette]
    modules: List[LoadedModule]
    liquids: List[Liquid]
    errors: List[ErrorOccurrence]



import argparse
from typing import Optional, Tuple, cast

from opentrons.hardware_control import API, Controller
from opentrons.hardware_control.types import HardwareFeatureFlags
from opentrons.drivers.smoothie_drivers import SmoothieDriver


def root_argparser(description: Optional[str] = None) -> argparse.ArgumentParser:
    parse = argparse.ArgumentParser(description=description)
    parse.add_argument(
        "-p", "--port", help="serial port of the smoothie", default="", type=str
    )
    return parse


async def build_driver(port: Optional[str] = None) -> Tuple[API, SmoothieDriver]:
    hardware = await API.build_hardware_controller(
        port=port, feature_flags=HardwareFeatureFlags.build_from_ff()
    )
    backend: Controller = cast(Controller, hardware._backend)
    return hardware, backend._smoothie_driver



"""Dispense-in-place command request, result, and implementation models."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from pydantic import Field

from .pipetting_common import (
    PipetteIdMixin,
    DispenseVolumeMixin,
    FlowRateMixin,
    BaseLiquidHandlingResult,
)
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from ..execution import PipettingHandler


DispenseInPlaceCommandType = Literal["dispenseInPlace"]


class DispenseInPlaceParams(PipetteIdMixin, DispenseVolumeMixin, FlowRateMixin):
    """Payload required to dispense in place."""

    pushOut: Optional[float] = Field(
        None,
        description="push the plunger a small amount farther than necessary for accurate low-volume dispensing",
    )


class DispenseInPlaceResult(BaseLiquidHandlingResult):
    """Result data from the execution of a DispenseInPlace command."""

    pass


class DispenseInPlaceImplementation(
    AbstractCommandImpl[DispenseInPlaceParams, DispenseInPlaceResult]
):
    """DispenseInPlace command implementation."""

    def __init__(self, pipetting: PipettingHandler, **kwargs: object) -> None:
        self._pipetting = pipetting

    async def execute(self, params: DispenseInPlaceParams) -> DispenseInPlaceResult:
        """Dispense without moving the pipette."""
        volume = await self._pipetting.dispense_in_place(
            pipette_id=params.pipetteId,
            volume=params.volume,
            flow_rate=params.flowRate,
            push_out=params.pushOut,
        )
        return DispenseInPlaceResult(volume=volume)


class DispenseInPlace(BaseCommand[DispenseInPlaceParams, DispenseInPlaceResult]):
    """DispenseInPlace command model."""

    commandType: DispenseInPlaceCommandType = "dispenseInPlace"
    params: DispenseInPlaceParams
    result: Optional[DispenseInPlaceResult]

    _ImplementationCls: Type[
        DispenseInPlaceImplementation
    ] = DispenseInPlaceImplementation


class DispenseInPlaceCreate(BaseCommandCreate[DispenseInPlaceParams]):
    """DispenseInPlace command request model."""

    commandType: DispenseInPlaceCommandType = "dispenseInPlace"
    params: DispenseInPlaceParams

    _CommandCls: Type[DispenseInPlace] = DispenseInPlace



"""Typing protocols describing a hardware controller."""
from typing_extensions import Protocol, Type

from opentrons.hardware_control.types import Axis

from .module_provider import ModuleProvider
from .hardware_manager import HardwareManager
from .chassis_accessory_manager import ChassisAccessoryManager
from .event_sourcer import EventSourcer
from .liquid_handler import LiquidHandler
from .calibratable import Calibratable
from .configurable import Configurable
from .motion_controller import MotionController
from .instrument_configurer import InstrumentConfigurer
from .execution_controllable import ExecutionControllable
from .asyncio_configurable import AsyncioConfigurable
from .stoppable import Stoppable
from .simulatable import Simulatable
from .identifiable import Identifiable
from .gripper_controller import GripperController
from .flex_calibratable import FlexCalibratable
from .flex_instrument_configurer import FlexInstrumentConfigurer

from .types import (
    CalibrationType,
    MountArgType,
    ConfigType,
    OT2RobotType,
    FlexRobotType,
)


class HardwareControlInterface(
    ModuleProvider,
    ExecutionControllable,
    LiquidHandler[CalibrationType, MountArgType, ConfigType],
    ChassisAccessoryManager,
    HardwareManager,
    AsyncioConfigurable,
    Stoppable,
    Simulatable,
    Identifiable[Type[OT2RobotType]],
    Protocol[CalibrationType, MountArgType, ConfigType],
):
    """A mypy protocol for a hardware controller.

    This class provides an protocol for the basic hardware controller class,
    with at least two implementations: one for the OT-2, and one for the
    OT-3. While the two classes have the same API, fundamental architectural
    decisions in the OT-2 hardware controller (specifically the data types used
    in the HardwareControl/backend split) make it unsuitable for the OT-3.

    This is a protocol rather than an ABC because of the use of wrapping adapters
    such as ThreadManager and SynchAdapter. Because those classes work via
    getattr, they can't inherit from an ABC that requires specific methods;
    however, they can satisfy protocols.
    """

    def get_robot_type(self) -> Type[OT2RobotType]:
        return OT2RobotType


class FlexHardwareControlInterface(
    ModuleProvider,
    ExecutionControllable,
    LiquidHandler[CalibrationType, MountArgType, ConfigType],
    ChassisAccessoryManager,
    HardwareManager,
    AsyncioConfigurable,
    Stoppable,
    Simulatable,
    GripperController,
    FlexCalibratable,
    FlexInstrumentConfigurer[MountArgType],
    Identifiable[Type[FlexRobotType]],
    Protocol[CalibrationType, MountArgType, ConfigType],
):
    """A mypy protocol for a hardware controller with Flex-specific extensions.

    The interface for the Flex controller is mostly in-line with the OT-2 interface,
    with some additional functionality and parameterization not supported on the OT-2.
    """

    def get_robot_type(self) -> Type[FlexRobotType]:
        return FlexRobotType

    def motor_status_ok(self, axis: Axis) -> bool:
        ...

    def encoder_status_ok(self, axis: Axis) -> bool:
        ...


__all__ = [
    "HardwareControlAPI",
    "FlexHardwareControlInterface",
    "Simulatable",
    "Stoppable",
    "AsyncioConfigurable",
    "ExecutionControllable",
    "InstrumentConfigurer",
    "MotionController",
    "Configurable",
    "Calibratable",
    "LiquidHandler",
    "EventSourcer",
    "ChassisAccessoryManager",
    "HardwareManager",
    "ModuleProvider",
    "Identifiable",
    "FlexCalibratable",
]



from .mod_abc import AbstractModule
from .tempdeck import TempDeck
from .magdeck import MagDeck
from .thermocycler import Thermocycler
from .heater_shaker import HeaterShaker
from .update import update_firmware
from .utils import MODULE_TYPE_BY_NAME, build
from .types import (
    ThermocyclerStep,
    UploadFunction,
    BundledFirmware,
    UpdateError,
    ModuleAtPort,
    ModuleType,
    ModuleModel,
    TemperatureStatus,
    MagneticStatus,
    HeaterShakerStatus,
    SpeedStatus,
    LiveData,
)

__all__ = [
    "MODULE_TYPE_BY_NAME",
    "build",
    "update_firmware",
    "ThermocyclerStep",
    "AbstractModule",
    "TempDeck",
    "MagDeck",
    "Thermocycler",
    "UploadFunction",
    "BundledFirmware",
    "UpdateError",
    "ModuleAtPort",
    "HeaterShaker",
    "ModuleType",
    "ModuleModel",
    "TemperatureStatus",
    "MagneticStatus",
    "HeaterShakerStatus",
    "SpeedStatus",
    "LiveData",
]



"""Blow-out in place command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal
from pydantic import BaseModel

from .pipetting_common import (
    PipetteIdMixin,
    FlowRateMixin,
)
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

from opentrons.hardware_control import HardwareControlAPI


if TYPE_CHECKING:
    from ..execution import PipettingHandler
    from ..state import StateView


BlowOutInPlaceCommandType = Literal["blowOutInPlace"]


class BlowOutInPlaceParams(PipetteIdMixin, FlowRateMixin):
    """Payload required to blow-out in place."""

    pass


class BlowOutInPlaceResult(BaseModel):
    """Result data from the execution of a BlowOutInPlace command."""

    pass


class BlowOutInPlaceImplementation(
    AbstractCommandImpl[BlowOutInPlaceParams, BlowOutInPlaceResult]
):
    """BlowOutInPlace command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        state_view: StateView,
        hardware_api: HardwareControlAPI,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api

    async def execute(self, params: BlowOutInPlaceParams) -> BlowOutInPlaceResult:
        """Blow-out without moving the pipette."""
        await self._pipetting.blow_out_in_place(
            pipette_id=params.pipetteId, flow_rate=params.flowRate
        )

        return BlowOutInPlaceResult()


class BlowOutInPlace(BaseCommand[BlowOutInPlaceParams, BlowOutInPlaceResult]):
    """BlowOutInPlace command model."""

    commandType: BlowOutInPlaceCommandType = "blowOutInPlace"
    params: BlowOutInPlaceParams
    result: Optional[BlowOutInPlaceResult]

    _ImplementationCls: Type[
        BlowOutInPlaceImplementation
    ] = BlowOutInPlaceImplementation


class BlowOutInPlaceCreate(BaseCommandCreate[BlowOutInPlaceParams]):
    """BlowOutInPlace command request model."""

    commandType: BlowOutInPlaceCommandType = "blowOutInPlace"
    params: BlowOutInPlaceParams

    _CommandCls: Type[BlowOutInPlace] = BlowOutInPlace



"""A helper for controlling a `ProtocolEngine` without async/await."""
from asyncio import AbstractEventLoop, run_coroutine_threadsafe
from typing import Any, overload
from typing_extensions import Literal

from opentrons_shared_data.labware.dev_types import LabwareUri
from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from ..protocol_engine import ProtocolEngine
from ..errors import ProtocolCommandFailedError
from ..state import StateView
from ..commands import CommandCreate, CommandResult


class ChildThreadTransport:
    """A helper for controlling a `ProtocolEngine` without async/await.

    You shouldn't use this directly, except to construct it and then pass it to a `SyncClient`.

    This class is responsible for doing the actual transformation from async `ProtocolEngine` calls
    to non-async ones, and doing it in a thread-safe way.
    """

    def __init__(self, engine: ProtocolEngine, loop: AbstractEventLoop) -> None:
        """Initialize the `ChildThreadTransport`.

        Args:
            engine: The `ProtocolEngine` instance that you want to interact with.
                It must be running in a thread *other* than the one from which you
                want to synchronously access it.
            loop: The event loop that `engine` is running in (in the other thread).
        """
        self._engine = engine
        self._loop = loop

    @property
    def state(self) -> StateView:
        """Get a view of the Protocol Engine's state."""
        return self._engine.state_view

    def execute_command(self, request: CommandCreate) -> CommandResult:
        """Execute a ProtocolEngine command, blocking until the command completes.

        Args:
            request: The ProtocolEngine command request

        Returns:
            The command's result data.

        Raises:
            ProtocolEngineError: if the command execution is not successful,
                the specific error that cause the command to fail is raised.
        """
        command = run_coroutine_threadsafe(
            self._engine.add_and_execute_command(request=request),
            loop=self._loop,
        ).result()

        # TODO: this needs to have an actual code
        if command.error is not None:
            error = command.error
            raise ProtocolCommandFailedError(
                original_error=error,
                message=f"{error.errorType}: {error.detail}",
            )

        # FIXME(mm, 2023-04-10): This assert can easily trigger from this sequence:
        #
        # 1. The engine is paused.
        # 2. The user's Python script calls this method to start a new command,
        #    which remains `queued` because of the pause.
        # 3. The engine is stopped.
        #
        # The returned command will be `queued`, so it won't have a result.
        #
        # We need to figure out a proper way to report this condition to callers
        # so they correctly interpret it as an intentional stop, not an internal error.
        assert command.result is not None, f"Expected Command {command} to have result"

        return command.result

    @overload
    def call_method(
        self,
        method_name: Literal["add_labware_definition"],
        *,
        definition: LabwareDefinition,
    ) -> LabwareUri:
        ...

    @overload
    def call_method(
        self,
        method_name: Literal["reset_tips"],
        *,
        labware_id: str,
    ) -> None:
        ...

    @overload
    def call_method(
        self,
        method_name: str,
        **kwargs: Any,
    ) -> Any:
        ...

    def call_method(self, method_name: str, **kwargs: Any) -> Any:
        """Execute a ProtocolEngine method, returning the result."""
        return run_coroutine_threadsafe(
            self._call_method(method_name, **kwargs),
            loop=self._loop,
        ).result()

    async def _call_method(self, method_name: str, **kwargs: Any) -> Any:
        method = getattr(self._engine, method_name)
        assert callable(method), f"{method_name} is not a method of ProtocolEngine"
        return method(**kwargs)



"""The interface that implements InstrumentContext."""

from __future__ import annotations

from abc import abstractmethod, ABC
from typing import Any, Generic, Optional, TypeVar, Union

from opentrons import types
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.protocols.api_support.util import FlowRates
from opentrons.protocol_api._nozzle_layout import NozzleLayout

from .._trash_bin import TrashBin
from .._waste_chute import WasteChute
from .well import WellCoreType


class AbstractInstrument(ABC, Generic[WellCoreType]):
    @abstractmethod
    def get_default_speed(self) -> float:
        ...

    @abstractmethod
    def set_default_speed(self, speed: float) -> None:
        ...

    @abstractmethod
    def aspirate(
        self,
        location: types.Location,
        well_core: Optional[WellCoreType],
        volume: float,
        rate: float,
        flow_rate: float,
        in_place: bool,
    ) -> None:
        """Aspirate a given volume of liquid from the specified location.
        Args:
            volume: The volume of liquid to aspirate, in microliters.
            location: The exact location to aspirate from.
            well_core: The well to aspirate from, if applicable.
            rate: The rate for how quickly to aspirate.
            flow_rate: The flow rate in µL/s to aspirate at.
            in_place: Whether this is in-place.
        """
        ...

    @abstractmethod
    def dispense(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[WellCoreType],
        volume: float,
        rate: float,
        flow_rate: float,
        in_place: bool,
        push_out: Optional[float],
    ) -> None:
        """Dispense a given volume of liquid into the specified location.
        Args:
            volume: The volume of liquid to dispense, in microliters.
            location: The exact location to dispense to.
            well_core: The well to dispense to, if applicable.
            rate: The rate for how quickly to dispense.
            flow_rate: The flow rate in µL/s to dispense at.
            in_place: Whether this is in-place.
            push_out: The amount to push the plunger below bottom position.
        """
        ...

    @abstractmethod
    def blow_out(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[WellCoreType],
        in_place: bool,
    ) -> None:
        """Blow liquid out of the tip.

        Args:
            location: The location to blow out into.
            well_core: The well to blow out into.
            in_place: Whether this is in-place.
        """
        ...

    @abstractmethod
    def touch_tip(
        self,
        location: types.Location,
        well_core: WellCoreType,
        radius: float,
        z_offset: float,
        speed: float,
    ) -> None:
        ...

    @abstractmethod
    def pick_up_tip(
        self,
        location: types.Location,
        well_core: WellCoreType,
        presses: Optional[int],
        increment: Optional[float],
        prep_after: bool = True,
    ) -> None:
        """Move to and pick up a tip from a given well.

        Args:
            location: The location of the well we're picking up from.
            well_core: The well to pick up from.
            presses: Customize the number of presses the pipette does.
            increment: Customize the movement "distance" of the pipette to press harder.
            prep_after: Move plunger to the "ready to aspirate" position after pick up.
        """
        ...

    @abstractmethod
    def drop_tip(
        self,
        location: Optional[types.Location],
        well_core: WellCoreType,
        home_after: Optional[bool],
        alternate_drop_location: Optional[bool] = False,
    ) -> None:
        """Move to and drop a tip into a given well.

        Args:
            location: The location of the well we're dropping to.
                If unspecified, the default drop location of the well will be used.
            well_core: The well we're dropping into
            home_after: Whether to home the pipette after the tip is dropped.
            alternate_drop_location: Whether to randomize the exact location to drop tip
                within the specified well.
        """
        ...

    @abstractmethod
    def drop_tip_in_disposal_location(
        self, disposal_location: Union[TrashBin, WasteChute], home_after: Optional[bool]
    ) -> None:
        """Move to and drop tip into a TrashBin or WasteChute.

        Args:
            disposal_location: The disposal location object we're dropping to.
            home_after: Whether to home the pipette after the tip is dropped.
        """
        ...

    @abstractmethod
    def home(self) -> None:
        ...

    @abstractmethod
    def home_plunger(self) -> None:
        ...

    @abstractmethod
    def move_to(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        well_core: Optional[WellCoreType],
        force_direct: bool,
        minimum_z_height: Optional[float],
        speed: Optional[float],
    ) -> None:
        ...

    @abstractmethod
    def get_mount(self) -> types.Mount:
        ...

    @abstractmethod
    def get_pipette_name(self) -> str:
        ...

    @abstractmethod
    def get_model(self) -> str:
        ...

    @abstractmethod
    def get_display_name(self) -> str:
        ...

    @abstractmethod
    def get_min_volume(self) -> float:
        ...

    @abstractmethod
    def get_max_volume(self) -> float:
        ...

    @abstractmethod
    def get_working_volume(self) -> float:
        ...

    @abstractmethod
    def get_current_volume(self) -> float:
        ...

    @abstractmethod
    def get_available_volume(self) -> float:
        ...

    @abstractmethod
    def get_hardware_state(self) -> PipetteDict:
        """Get the current state of the pipette hardware as a dictionary."""
        ...

    @abstractmethod
    def get_channels(self) -> int:
        ...

    @abstractmethod
    def get_active_channels(self) -> int:
        ...

    @abstractmethod
    def has_tip(self) -> bool:
        ...

    @abstractmethod
    def get_return_height(self) -> float:
        ...

    @abstractmethod
    def get_flow_rate(self) -> FlowRates:
        ...

    @abstractmethod
    def get_aspirate_flow_rate(self, rate: float = 1.0) -> float:
        ...

    @abstractmethod
    def get_dispense_flow_rate(self, rate: float = 1.0) -> float:
        ...

    @abstractmethod
    def get_blow_out_flow_rate(self, rate: float = 1.0) -> float:
        ...

    @abstractmethod
    def set_flow_rate(
        self,
        aspirate: Optional[float] = None,
        dispense: Optional[float] = None,
        blow_out: Optional[float] = None,
    ) -> None:
        ...

    @abstractmethod
    def configure_for_volume(self, volume: float) -> None:
        """Configure the pipette for a specific volume.

        Args:
            volume: The volume to prepare to handle.
        """
        ...

    @abstractmethod
    def prepare_to_aspirate(self) -> None:
        """Prepare the pipette to aspirate."""
        ...

    @abstractmethod
    def configure_nozzle_layout(
        self,
        style: NozzleLayout,
        primary_nozzle: Optional[str],
        front_right_nozzle: Optional[str],
    ) -> None:
        """Configure the pipette to a specific nozzle layout.

        Args:
            style: The type of configuration you wish to build.
            primary_nozzle: The nozzle that will determine a pipette's critical point.
            front_right_nozzle: The front right most nozzle in the requested layout.
        """
        ...

    @abstractmethod
    def is_tip_tracking_available(self) -> bool:
        """Return whether auto tip tracking is available for the pipette's current nozzle configuration."""
        ...


InstrumentCoreType = TypeVar("InstrumentCoreType", bound=AbstractInstrument[Any])



from __future__ import annotations

import logging
from contextlib import ExitStack
from typing import Any, List, Optional, Sequence, Union, cast, Dict
from opentrons_shared_data.errors.exceptions import (
    CommandPreconditionViolated,
    CommandParameterLimitViolated,
    UnexpectedTipRemovalError,
)
from opentrons.legacy_broker import LegacyBroker
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons import types
from opentrons.commands import commands as cmds

from opentrons.commands import publisher
from opentrons.protocols.advanced_control.mix import mix_from_kwargs
from opentrons.protocols.advanced_control import transfers

from opentrons.protocols.api_support.deck_type import NoTrashDefinedError
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support import instrument
from opentrons.protocols.api_support.util import (
    FlowRates,
    PlungerSpeeds,
    clamp_value,
    requires_version,
    APIVersionError,
)

from .core.common import InstrumentCore, ProtocolCore
from .core.engine import ENGINE_CORE_API_VERSION
from .core.legacy.legacy_instrument_core import LegacyInstrumentCore
from .config import Clearances
from ._trash_bin import TrashBin
from ._waste_chute import WasteChute
from ._nozzle_layout import NozzleLayout
from . import labware, validation


AdvancedLiquidHandling = Union[
    labware.Well,
    types.Location,
    Sequence[Union[labware.Well, types.Location]],
    Sequence[Sequence[labware.Well]],
]

_DEFAULT_ASPIRATE_CLEARANCE = 1.0
_DEFAULT_DISPENSE_CLEARANCE = 1.0

_log = logging.getLogger(__name__)

_PREP_AFTER_ADDED_IN = APIVersion(2, 13)
"""The version after which the pick-up tip procedure should also prepare the plunger."""
_PRESSES_INCREMENT_REMOVED_IN = APIVersion(2, 14)
"""The version after which the pick-up tip procedure deprecates presses and increment arguments."""
_DROP_TIP_LOCATION_ALTERNATING_ADDED_IN = APIVersion(2, 15)
"""The version after which a drop-tip-into-trash procedure drops tips in different alternating locations within the trash well."""
_PARTIAL_NOZZLE_CONFIGURATION_ADDED_IN = APIVersion(2, 16)


class InstrumentContext(publisher.CommandPublisher):
    """
    A context for a specific pipette or instrument.

    The InstrumentContext class provides the objects, attributes, and methods that allow
    you to use pipettes in your protocols.

    Methods generally fall into one of two categories.

      - They can change the state of the InstrumentContext object, like how fast it
        moves liquid or where it disposes of used tips.

      - They can command the instrument to perform an action, like picking up tips,
        moving to certain locations, and aspirating or dispensing liquid.

    Objects in this class should not be instantiated directly. Instead, instances are
    returned by :py:meth:`ProtocolContext.load_instrument`.

    .. versionadded:: 2.0

    """

    def __init__(
        self,
        core: InstrumentCore,
        protocol_core: ProtocolCore,
        broker: LegacyBroker,
        api_version: APIVersion,
        tip_racks: List[labware.Labware],
        trash: Optional[Union[labware.Labware, TrashBin, WasteChute]],
        requested_as: str,
    ) -> None:
        super().__init__(broker)
        self._api_version = api_version
        self._core = core
        self._protocol_core = protocol_core
        self._tip_racks = tip_racks
        self._last_tip_picked_up_from: Union[labware.Well, None] = None
        self._starting_tip: Union[labware.Well, None] = None
        self._well_bottom_clearances = Clearances(
            default_aspirate=_DEFAULT_ASPIRATE_CLEARANCE,
            default_dispense=_DEFAULT_DISPENSE_CLEARANCE,
        )
        self._user_specified_trash: Union[
            labware.Labware, TrashBin, WasteChute, None
        ] = trash
        self.requested_as = requested_as

    @property
    @requires_version(2, 0)
    def api_version(self) -> APIVersion:
        return self._api_version

    @property
    @requires_version(2, 0)
    def starting_tip(self) -> Union[labware.Well, None]:
        """
        Which well of a tip rack the pipette should start at when automatically choosing tips to pick up.

        See :py:meth:`.pick_up_tip()`.

        .. note::

            In robot software versions 6.3.0 and 6.3.1, protocols specifying API level
            2.14 ignored ``starting_tip`` on the second and subsequent calls to
            :py:meth:`.InstrumentContext.pick_up_tip` with no argument. This is fixed
            for all API levels as of robot software version 7.0.0.
        """
        return self._starting_tip

    @starting_tip.setter
    def starting_tip(self, location: Union[labware.Well, None]) -> None:
        self._starting_tip = location

    @requires_version(2, 0)
    def reset_tipracks(self) -> None:
        """Reload all tips in each tip rack and reset the starting tip."""
        for tiprack in self.tip_racks:
            tiprack.reset()
        self.starting_tip = None

    @property
    @requires_version(2, 0)
    def default_speed(self) -> float:
        """The speed at which the robot's gantry moves in mm/s.

        The default speed for Flex varies between 300 and 350 mm/s. The OT-2 default is
        400 mm/s. In addition to changing the default, the speed of individual motions
        can be changed with the ``speed`` argument of the
        :py:meth:`InstrumentContext.move_to` method. See :ref:`gantry_speed`.
        """
        return self._core.get_default_speed()

    @default_speed.setter
    def default_speed(self, speed: float) -> None:
        self._core.set_default_speed(speed)

    @requires_version(2, 0)
    def aspirate(
        self,
        volume: Optional[float] = None,
        location: Optional[Union[types.Location, labware.Well]] = None,
        rate: float = 1.0,
    ) -> InstrumentContext:
        """
        Draw liquid into a pipette tip.

        See :ref:`new-aspirate` for more details and examples.

        :param volume: The volume to aspirate, measured in µL. If unspecified,
                    defaults to the maximum volume for the pipette and its currently
                    attached tip.

                    If ``aspirate`` is called with a volume of precisely 0, its behavior
                    depends on the API level of the protocol. On API levels below 2.16,
                    it will behave the same as a volume of ``None``/unspecified: aspirate
                    until the pipette is full. On API levels at or above 2.16, no liquid
                    will be aspirated.
        :type volume: int or float
        :param location: Tells the robot where to aspirate from. The location can be
                         a :py:class:`.Well` or a :py:class:`.Location`.

                            - If the location is a ``Well``, the robot will aspirate at
                              or above the bottom center of the well. The distance (in mm)
                              from the well bottom is specified by
                              :py:obj:`well_bottom_clearance.aspirate
                              <well_bottom_clearance>`.

                            - If the location is a ``Location`` (e.g., the result of
                              :py:meth:`.Well.top` or :py:meth:`.Well.bottom`), the robot
                              will aspirate from that specified position.

                            - If the ``location`` is unspecified, the robot will
                              aspirate from its current position.
        :param rate: A multiplier for the default flow rate of the pipette. Calculated
                     as ``rate`` multiplied by :py:attr:`flow_rate.aspirate
                     <flow_rate>`. If not specified, defaults to 1.0. See
                     :ref:`new-plunger-flow-rates`.
        :type rate: float
        :returns: This instance.

        .. note::

            If ``aspirate`` is called with a single, unnamed argument, it will treat
            that argument as ``volume``. If you want to call ``aspirate`` with only
            ``location``, specify it as a keyword argument:
            ``pipette.aspirate(location=plate['A1'])``

        """
        _log.debug(
            "aspirate {} from {} at {}".format(
                volume, location if location else "current position", rate
            )
        )

        well: Optional[labware.Well] = None
        move_to_location: types.Location

        last_location = self._get_last_location_by_api_version()
        try:
            target = validation.validate_location(
                location=location, last_location=last_location
            )
        except validation.NoLocationError as e:
            raise RuntimeError(
                "If aspirate is called without an explicit location, another"
                " method that moves to a location (such as move_to or "
                "dispense) must previously have been called so the robot "
                "knows where it is."
            ) from e

        if isinstance(target, validation.WellTarget):
            move_to_location = target.location or target.well.bottom(
                z=self._well_bottom_clearances.aspirate
            )
            well = target.well
        if isinstance(target, validation.PointTarget):
            move_to_location = target.location
        if isinstance(target, (TrashBin, WasteChute)):
            raise ValueError(
                "Trash Bin and Waste Chute are not acceptable location parameters for Aspirate commands."
            )
        if self.api_version >= APIVersion(2, 11):
            instrument.validate_takes_liquid(
                location=move_to_location,
                reject_module=self.api_version >= APIVersion(2, 13),
                reject_adapter=self.api_version >= APIVersion(2, 15),
            )

        if self.api_version >= APIVersion(2, 16):
            c_vol = self._core.get_available_volume() if volume is None else volume
        else:
            c_vol = self._core.get_available_volume() if not volume else volume
        flow_rate = self._core.get_aspirate_flow_rate(rate)

        with publisher.publish_context(
            broker=self.broker,
            command=cmds.aspirate(
                instrument=self,
                volume=c_vol,
                location=move_to_location,
                flow_rate=flow_rate,
                rate=rate,
            ),
        ):
            self._core.aspirate(
                location=move_to_location,
                well_core=well._core if well is not None else None,
                volume=c_vol,
                rate=rate,
                flow_rate=flow_rate,
                in_place=target.in_place,
            )

        return self

    @requires_version(2, 0)
    def dispense(  # noqa: C901
        self,
        volume: Optional[float] = None,
        location: Optional[
            Union[types.Location, labware.Well, TrashBin, WasteChute]
        ] = None,
        rate: float = 1.0,
        push_out: Optional[float] = None,
    ) -> InstrumentContext:
        """
        Dispense liquid from a pipette tip.

        See :ref:`new-dispense` for more details and examples.

        :param volume: The volume to dispense, measured in µL. If unspecified,
                       defaults to :py:attr:`current_volume`. If only a volume is
                       passed, the pipette will dispense from its current position.

                       If ``dispense`` is called with a volume of precisely 0, its behavior
                       depends on the API level of the protocol. On API levels below 2.16,
                       it will behave the same as a volume of ``None``/unspecified: dispense
                       all liquid in the pipette. On API levels at or above 2.16, no liquid
                       will be dispensed.
        :type volume: int or float

        :param location: Tells the robot where to dispense liquid held in the pipette.
                         The location can be a :py:class:`.Well` or a
                         :py:class:`.Location`.

                            - If the location is a ``Well``, the pipette will dispense
                              at or above the bottom center of the well. The distance (in
                              mm) from the well bottom is specified by
                              :py:obj:`well_bottom_clearance.dispense
                              <well_bottom_clearance>`.

                            - If the location is a ``Location`` (e.g., the result of
                              :py:meth:`.Well.top` or :py:meth:`.Well.bottom`), the robot
                              will dispense into that specified position.

                            - If the ``location`` is unspecified, the robot will
                              dispense into its current position.

                            If only a ``location`` is passed (e.g.,
                            ``pipette.dispense(location=plate['A1'])``), all of the
                            liquid aspirated into the pipette will be dispensed (the
                            amount is accessible through :py:attr:`current_volume`).

        :param rate: How quickly a pipette dispenses liquid. The speed in µL/s is
                     calculated as ``rate`` multiplied by :py:attr:`flow_rate.dispense
                     <flow_rate>`. If not specified, defaults to 1.0. See
                     :ref:`new-plunger-flow-rates`.
        :type rate: float
        :param push_out: Continue past the plunger bottom to help ensure all liquid
                         leaves the tip. Measured in µL. The default value is ``None``.

                         See :ref:`push-out-dispense` for details.
        :type push_out: float

        :returns: This instance.

        .. note::

            If ``dispense`` is called with a single, unnamed argument, it will treat
            that argument as ``volume``. If you want to call ``dispense`` with only
            ``location``, specify it as a keyword argument:
            ``pipette.dispense(location=plate['A1'])``.

        .. versionchanged:: 2.15
            Added the ``push_out`` parameter.
        """
        if self.api_version < APIVersion(2, 15) and push_out:
            raise APIVersionError(
                "Unsupported parameter push_out. Change your API version to 2.15 or above to use this parameter."
            )
        _log.debug(
            "dispense {} from {} at {}".format(
                volume, location if location else "current position", rate
            )
        )
        well: Optional[labware.Well] = None
        last_location = self._get_last_location_by_api_version()

        try:
            target = validation.validate_location(
                location=location, last_location=last_location
            )
        except validation.NoLocationError as e:
            raise RuntimeError(
                "If dispense is called without an explicit location, another"
                " method that moves to a location (such as move_to or "
                "aspirate) must previously have been called so the robot "
                "knows where it is."
            ) from e

        if isinstance(target, validation.WellTarget):
            well = target.well
            if target.location:
                move_to_location = target.location
            elif well.parent._core.is_fixed_trash():
                move_to_location = target.well.top()
            else:
                move_to_location = target.well.bottom(
                    z=self._well_bottom_clearances.dispense
                )
        if isinstance(target, validation.PointTarget):
            move_to_location = target.location

        if self.api_version >= APIVersion(2, 11) and not isinstance(
            target, (TrashBin, WasteChute)
        ):
            instrument.validate_takes_liquid(
                location=move_to_location,
                reject_module=self.api_version >= APIVersion(2, 13),
                reject_adapter=self.api_version >= APIVersion(2, 15),
            )

        if self.api_version >= APIVersion(2, 16):
            c_vol = self._core.get_current_volume() if volume is None else volume
        else:
            c_vol = self._core.get_current_volume() if not volume else volume

        flow_rate = self._core.get_dispense_flow_rate(rate)

        if isinstance(target, (TrashBin, WasteChute)):
            with publisher.publish_context(
                broker=self.broker,
                command=cmds.dispense_in_disposal_location(
                    instrument=self,
                    volume=c_vol,
                    location=target,
                    rate=rate,
                    flow_rate=flow_rate,
                ),
            ):
                self._core.dispense(
                    volume=c_vol,
                    rate=rate,
                    location=target,
                    well_core=None,
                    flow_rate=flow_rate,
                    in_place=False,
                    push_out=push_out,
                )
            return self

        with publisher.publish_context(
            broker=self.broker,
            command=cmds.dispense(
                instrument=self,
                volume=c_vol,
                location=move_to_location,
                rate=rate,
                flow_rate=flow_rate,
            ),
        ):
            self._core.dispense(
                volume=c_vol,
                rate=rate,
                location=move_to_location,
                well_core=well._core if well is not None else None,
                flow_rate=flow_rate,
                in_place=target.in_place,
                push_out=push_out,
            )

        return self

    @requires_version(2, 0)
    def mix(
        self,
        repetitions: int = 1,
        volume: Optional[float] = None,
        location: Optional[Union[types.Location, labware.Well]] = None,
        rate: float = 1.0,
    ) -> InstrumentContext:
        """
        Mix a volume of liquid by repeatedly aspirating and dispensing it in a single location.

        See :ref:`mix` for examples.

        :param repetitions: Number of times to mix (default is 1).
        :param volume: The volume to mix, measured in µL. If unspecified, defaults
                       to the maximum volume for the pipette and its attached tip.

                       If ``mix`` is called with a volume of precisely 0, its behavior
                       depends on the API level of the protocol. On API levels below 2.16,
                       it will behave the same as a volume of ``None``/unspecified: mix
                       the full working volume of the pipette. On API levels at or above 2.16,
                       no liquid will be mixed.
        :param location: The :py:class:`.Well` or :py:class:`~.types.Location` where the
                        pipette will mix. If unspecified, the pipette will mix at its
                        current position.
        :param rate: How quickly the pipette aspirates and dispenses liquid while
                     mixing. The aspiration flow rate is calculated as ``rate``
                     multiplied by :py:attr:`flow_rate.aspirate <flow_rate>`. The
                     dispensing flow rate is calculated as ``rate`` multiplied by
                     :py:attr:`flow_rate.dispense <flow_rate>`. See
                     :ref:`new-plunger-flow-rates`.
        :raises: ``UnexpectedTipRemovalError`` -- If no tip is attached to the pipette.
        :returns: This instance.

        .. note::

            All the arguments of ``mix`` are optional. However, if you omit one of them,
            all subsequent arguments must be passed as keyword arguments. For instance,
            ``pipette.mix(1, location=wellplate['A1'])`` is a valid call, but
            ``pipette.mix(1, wellplate['A1'])`` is not.

        """
        _log.debug(
            "mixing {}uL with {} repetitions in {} at rate={}".format(
                volume, repetitions, location if location else "current position", rate
            )
        )
        if not self._core.has_tip():
            raise UnexpectedTipRemovalError("mix", self.name, self.mount)

        if self.api_version >= APIVersion(2, 16):
            c_vol = self._core.get_available_volume() if volume is None else volume
        else:
            c_vol = self._core.get_available_volume() if not volume else volume

        dispense_kwargs: Dict[str, Any] = {}
        if self.api_version >= APIVersion(2, 16):
            dispense_kwargs["push_out"] = 0.0

        with publisher.publish_context(
            broker=self.broker,
            command=cmds.mix(
                instrument=self,
                repetitions=repetitions,
                volume=c_vol,
                location=location,
            ),
        ):
            self.aspirate(volume, location, rate)
            while repetitions - 1 > 0:
                self.dispense(volume, rate=rate, **dispense_kwargs)
                self.aspirate(volume, rate=rate)
                repetitions -= 1
            self.dispense(volume, rate=rate)

        return self

    @requires_version(2, 0)
    def blow_out(
        self,
        location: Optional[
            Union[types.Location, labware.Well, TrashBin, WasteChute]
        ] = None,
    ) -> InstrumentContext:
        """
        Blow an extra amount of air through a pipette's tip to clear it.

        If :py:meth:`dispense` is used to empty a pipette, usually a small amount of
        liquid remains in the tip. During a blowout, the pipette moves the plunger
        beyond its normal limits to help remove all liquid from the pipette tip. See
        :ref:`blow-out`.

        :param location: The blowout location. If no location is specified, the pipette
                         will blow out from its current position.
        :type location: :py:class:`.Well` or :py:class:`.Location` or ``None``

        :raises RuntimeError: If no location is specified and the location cache is
                              ``None``. This should happen if ``blow_out()`` is called
                              without first calling a method that takes a location, like
                              :py:meth:`.aspirate` or :py:meth:`dispense`.
        :returns: This instance.
        """
        well: Optional[labware.Well] = None
        move_to_location: types.Location

        last_location = self._get_last_location_by_api_version()
        try:
            target = validation.validate_location(
                location=location, last_location=last_location
            )
        except validation.NoLocationError as e:
            raise RuntimeError(
                "If blow out is called without an explicit location, another"
                " method that moves to a location (such as move_to or "
                "dispense) must previously have been called so the robot "
                "knows where it is."
            ) from e

        if isinstance(target, validation.WellTarget):
            if target.well.parent.is_tiprack:
                _log.warning(
                    "Blow_out being performed on a tiprack. "
                    "Please re-check your code"
                )
            move_to_location = target.location or target.well.top()
            well = target.well
        elif isinstance(target, validation.PointTarget):
            move_to_location = target.location
        elif isinstance(target, (TrashBin, WasteChute)):
            with publisher.publish_context(
                broker=self.broker,
                command=cmds.blow_out_in_disposal_location(
                    instrument=self, location=target
                ),
            ):
                self._core.blow_out(
                    location=target,
                    well_core=None,
                    in_place=False,
                )
            return self

        with publisher.publish_context(
            broker=self.broker,
            command=cmds.blow_out(instrument=self, location=move_to_location),
        ):
            self._core.blow_out(
                location=move_to_location,
                well_core=well._core if well is not None else None,
                in_place=target.in_place,
            )

        return self

    def _determine_speed(self, speed: float) -> float:
        if self.api_version < APIVersion(2, 4):
            return clamp_value(speed, 80, 20, "touch_tip:")
        else:
            return clamp_value(speed, 80, 1, "touch_tip:")

    @publisher.publish(command=cmds.touch_tip)
    @requires_version(2, 0)
    def touch_tip(
        self,
        location: Optional[labware.Well] = None,
        radius: float = 1.0,
        v_offset: float = -1.0,
        speed: float = 60.0,
    ) -> InstrumentContext:
        """
        Touch the pipette tip to the sides of a well, with the intent of removing leftover droplets.

        See :ref:`touch-tip` for more details and examples.

        :param location: If no location is passed, the pipette will touch its tip at the
                         edges of the current well.
        :type location: :py:class:`.Well` or ``None``
        :param radius: How far to move, as a proportion of the target well's radius.
                       When ``radius=1.0``, the pipette tip will move all the way to the
                       edge of the target well. When ``radius=0.5``, it will move to 50%
                       of the well's radius. Default is 1.0 (100%)
        :type radius: float
        :param v_offset: How far above or below the well to touch the tip, measured in mm.
                         A positive offset moves the tip higher above the well.
                         A negative offset moves the tip lower into the well.
                         Default is -1.0 mm.
        :type v_offset: float
        :param speed: The speed for touch tip motion, in mm/s.

                        - Default: 60.0 mm/s
                        - Maximum: 80.0 mm/s
                        - Minimum: 1.0 mm/s
        :type speed: float
        :raises: ``UnexpectedTipRemovalError`` -- If no tip is attached to the pipette.
        :raises RuntimeError: If no location is specified and the location cache is
                              ``None``. This should happen if ``touch_tip`` is called
                              without first calling a method that takes a location, like
                              :py:meth:`.aspirate` or :py:meth:`dispense`.
        :returns: This instance.
        """
        if not self._core.has_tip():
            raise UnexpectedTipRemovalError("touch_tip", self.name, self.mount)

        checked_speed = self._determine_speed(speed)

        # If location is a valid well, move to the well first
        if location is None:
            last_location = self._protocol_core.get_last_location()
            if not last_location:
                raise RuntimeError("No valid current location cache present")
            parent_labware, well = last_location.labware.get_parent_labware_and_well()
            if not well or not parent_labware:
                raise RuntimeError(
                    f"Last location {location} has no associated well or labware."
                )
        elif isinstance(location, labware.Well):
            well = location
            parent_labware = well.parent
        else:
            raise TypeError(f"location should be a Well, but it is {location}")

        if "touchTipDisabled" in parent_labware.quirks:
            _log.info(f"Ignoring touch tip on labware {well}")
            return self
        if parent_labware.is_tiprack:
            _log.warning(
                "Touch_tip being performed on a tiprack. Please re-check your code"
            )

        if self.api_version < APIVersion(2, 4):
            move_to_location = well.top()
        else:
            move_to_location = well.top(z=v_offset)

        self._core.touch_tip(
            location=move_to_location,
            well_core=well._core,
            radius=radius,
            z_offset=v_offset,
            speed=checked_speed,
        )
        return self

    @publisher.publish(command=cmds.air_gap)
    @requires_version(2, 0)
    def air_gap(
        self, volume: Optional[float] = None, height: Optional[float] = None
    ) -> InstrumentContext:
        """
        Draw air into the pipette's tip at the current well.

        See :ref:`air-gap`.

        :param volume: The amount of air, measured in µL. Calling ``air_gap()`` with no
                       arguments uses the entire remaining volume in the pipette.
        :type volume: float

        :param height: The height, in mm, to move above the current well before creating
                       the air gap. The default is 5 mm above the current well.
        :type height: float

        :raises: ``UnexpectedTipRemovalError`` -- If no tip is attached to the pipette.

        :raises RuntimeError: If location cache is ``None``. This should happen if
                              ``air_gap()`` is called without first calling a method
                              that takes a location (e.g., :py:meth:`.aspirate`,
                              :py:meth:`dispense`)

        :returns: This instance.

        .. note::

            Both ``volume`` and ``height`` are optional, but if you want to specify only
            ``height`` you must do it as a keyword argument:
            ``pipette.air_gap(height=2)``. If you call ``air_gap`` with a single,
            unnamed argument, it will always be interpreted as a volume.


        """
        if not self._core.has_tip():
            raise UnexpectedTipRemovalError("air_gap", self.name, self.mount)

        if height is None:
            height = 5
        loc = self._protocol_core.get_last_location()
        if not loc or not loc.labware.is_well:
            raise RuntimeError("No previous Well cached to perform air gap")
        target = loc.labware.as_well().top(height)
        self.move_to(target, publish=False)
        self.aspirate(volume)
        return self

    @publisher.publish(command=cmds.return_tip)
    @requires_version(2, 0)
    def return_tip(self, home_after: Optional[bool] = None) -> InstrumentContext:
        """
        Drop the currently attached tip in its original location in the tip rack.

        Returning a tip does not reset tip tracking, so :py:obj:`.Well.has_tip` will
        remain ``False`` for the destination.

        :returns: This instance.

        :param home_after: See the ``home_after`` parameter of :py:meth:`drop_tip`.
        """
        if not self._core.has_tip():
            _log.warning("Pipette has no tip to return")

        loc = self._last_tip_picked_up_from

        if not isinstance(loc, labware.Well):
            raise TypeError(f"Last tip location should be a Well but it is: {loc}")

        self.drop_tip(loc, home_after=home_after)

        return self

    @requires_version(2, 0)
    def pick_up_tip(  # noqa: C901
        self,
        location: Union[types.Location, labware.Well, labware.Labware, None] = None,
        presses: Optional[int] = None,
        increment: Optional[float] = None,
        prep_after: Optional[bool] = None,
    ) -> InstrumentContext:
        """
        Pick up a tip for the pipette to run liquid-handling commands.

        See :ref:`basic-tip-pickup`.

        If no location is passed, the pipette will pick up the next available tip in its
        :py:attr:`~.InstrumentContext.tip_racks` list. Within each tip rack, tips will
        be picked up in the order specified by the labware definition and
        :py:meth:`.Labware.wells`. To adjust where the sequence starts, use
        :py:obj:`.starting_tip`.

        :param location: The location from which to pick up a tip. The ``location``
                         argument can be specified in several ways:

                           * As a :py:class:`.Well`. For example,
                             ``pipette.pick_up_tip(tiprack.wells()[0])`` will always pick
                             up the first tip in ``tiprack``, even if the rack is not a
                             member of :py:obj:`.InstrumentContext.tip_racks`.

                           * As a labware. ``pipette.pick_up_tip(tiprack)`` will pick up
                             the next available tip in ``tiprack``, even if the rack is
                             not a member of :py:obj:`.InstrumentContext.tip_racks`.

                           * As a :py:class:`~.types.Location`. Use this to make fine
                             adjustments to the pickup location. For example, to tell
                             the robot to start its pick up tip routine 1 mm closer to
                             the top of the well in the tip rack, call
                             ``pipette.pick_up_tip(tiprack["A1"].top(z=-1))``.
        :type location: :py:class:`.Well` or :py:class:`.Labware` or :py:class:`.types.Location`
        :param presses: The number of times to lower and then raise the pipette when
                        picking up a tip, to ensure a good seal. Zero (``0``) will
                        result in the pipette hovering over the tip but not picking it
                        up (generally not desirable, but could be used for a dry run).

                            .. deprecated:: 2.14
                                Use the Opentrons App to change pipette pick-up settings.
        :type presses: int
        :param increment: The additional distance to travel on each successive press.
                          For example, if ``presses=3`` and ``increment=1.0``, then the
                          first press will travel down into the tip by 3.5 mm, the
                          second by 4.5 mm, and the third by 5.5 mm).

                              .. deprecated:: 2.14
                                  Use the Opentrons App to change pipette pick-up settings.
        :type increment: float
        :param prep_after: Whether the pipette plunger should prepare itself to aspirate
                           immediately after picking up a tip.

                           If ``True``, the pipette will move its plunger position to
                           bottom in preparation for any following calls to
                           :py:meth:`.aspirate`.

                           If ``False``, the pipette will prepare its plunger later,
                           during the next call to :py:meth:`.aspirate`. This is
                           accomplished by moving the tip to the top of the well, and
                           positioning the plunger outside any potential liquids.

                           .. warning::
                               This is provided for compatibility with older Python
                               Protocol API behavior. You should normally leave this
                               unset.

                               Setting ``prep_after=False`` may create an unintended
                               pipette movement, when the pipette automatically moves
                               the tip to the top of the well to prepare the plunger.
        :type prep_after: bool

        .. versionchanged:: 2.13
            Adds the ``prep_after`` argument. In version 2.12 and earlier, the plunger
            can't prepare itself for aspiration during :py:meth:`.pick_up_tip`, and will
            instead always prepare during :py:meth:`.aspirate`. Version 2.12 and earlier
            will raise an ``APIVersionError`` if a value is set for ``prep_after``.

        :returns: This instance.
        """

        if presses is not None and self._api_version >= _PRESSES_INCREMENT_REMOVED_IN:
            raise APIVersionError(
                f"presses is only available in API versions lower than {_PRESSES_INCREMENT_REMOVED_IN},"
                f" but you are using API {self._api_version}."
            )

        if increment is not None and self._api_version >= _PRESSES_INCREMENT_REMOVED_IN:
            raise APIVersionError(
                f"increment is only available in API versions lower than {_PRESSES_INCREMENT_REMOVED_IN},"
                f" but you are using API {self._api_version}."
            )

        if prep_after is not None and self._api_version < _PREP_AFTER_ADDED_IN:
            raise APIVersionError(
                f"prep_after is only available in API {_PREP_AFTER_ADDED_IN} and newer,"
                f" but you are using API {self._api_version}."
            )

        well: labware.Well
        tip_rack: labware.Labware
        move_to_location: Optional[types.Location] = None
        active_channels = (
            self.active_channels
            if self._api_version >= _PARTIAL_NOZZLE_CONFIGURATION_ADDED_IN
            else self.channels
        )

        if location is None:
            if not self._core.is_tip_tracking_available():
                raise CommandPreconditionViolated(
                    "Automatic tip tracking is not available for the current pipette"
                    " nozzle configuration. We suggest switching to a configuration"
                    " that supports automatic tip tracking or specifying the exact tip"
                    " to pick up."
                )

            tip_rack, well = labware.next_available_tip(
                starting_tip=self.starting_tip,
                tip_racks=self.tip_racks,
                channels=active_channels,
            )

        elif isinstance(location, labware.Well):
            well = location
            tip_rack = well.parent

        elif isinstance(location, labware.Labware):
            tip_rack, well = labware.next_available_tip(
                starting_tip=None,
                tip_racks=[location],
                channels=active_channels,
            )

        elif isinstance(location, types.Location):
            maybe_tip_rack, maybe_well = location.labware.get_parent_labware_and_well()

            if maybe_well is not None:
                well = maybe_well
                tip_rack = well.parent
                move_to_location = location

            elif maybe_tip_rack is not None:
                tip_rack, well = labware.next_available_tip(
                    starting_tip=None,
                    tip_racks=[maybe_tip_rack],
                    channels=active_channels,
                )
            else:
                raise TypeError(
                    "If specified as a `types.Location`,"
                    " `location` should refer to a ``Labware` or `Well` location."
                    f" However, it refers to {location.labware}"
                )

        else:
            raise TypeError(
                "If specified, location should be an instance of"
                " `types.Location` (e.g. the return value from `Well.top()`),"
                "  `Labware` or `Well` (e.g. `tiprack.wells()[0]`)."
                f" However, it is {location}"
            )

        instrument.validate_tiprack(self.name, tip_rack, _log)

        move_to_location = move_to_location or well.top()
        prep_after = (
            prep_after
            if prep_after is not None
            else self.api_version >= _PREP_AFTER_ADDED_IN
        )

        with publisher.publish_context(
            broker=self.broker,
            command=cmds.pick_up_tip(instrument=self, location=well),
        ):
            self._core.pick_up_tip(
                location=move_to_location,
                well_core=well._core,
                presses=presses,
                increment=increment,
                prep_after=prep_after,
            )

        self._last_tip_picked_up_from = well

        return self

    @requires_version(2, 0)
    def drop_tip(
        self,
        location: Optional[
            Union[
                types.Location,
                labware.Well,
                TrashBin,
                WasteChute,
            ]
        ] = None,
        home_after: Optional[bool] = None,
    ) -> InstrumentContext:
        """
        Drop the current tip.

        See :ref:`pipette-drop-tip` for examples.

        If no location is passed (e.g. ``pipette.drop_tip()``), the pipette will drop
        the attached tip into its :py:attr:`trash_container`.

        Starting with API version 2.15, if the trash container is the default fixed
        trash, the API will instruct the pipette to drop tips in different locations
        within the trash container. Varying the tip drop location helps prevent tips
        from piling up in a single location.

        The location in which to drop the tip can be manually specified with the
        ``location`` argument. The ``location`` argument can be specified in several
        ways:

            - As a :py:class:`.Well`. This uses a default location relative to the well.
              This style of call can be used to make the robot drop a tip into labware
              like a well plate or a reservoir. For example,
              ``pipette.drop_tip(location=reservoir["A1"])``.
            - As a :py:class:`~.types.Location`. For example, to drop a tip from an
              unusually large height above the tip rack, you could call
              ``pipette.drop_tip(tip_rack["A1"].top(z=10))``.
            - As a :py:class:`.TrashBin`. This uses a default location relative to the
              ``TrashBin`` object. For example,
              ``pipette.drop_tip(location=trash_bin)``.
            - As a :py:class:`.WasteChute`. This uses a default location relative to
              the ``WasteChute`` object. For example,
              ``pipette.drop_tip(location=waste_chute)``.

        :param location:
            The location to drop the tip.
        :type location:
            :py:class:`~.types.Location` or :py:class:`.Well` or ``None``
        :param home_after:
            Whether to home the pipette's plunger after dropping the tip. If not
            specified, defaults to ``True`` on an OT-2.

            When ``False``, the pipette does not home its plunger. This can save a few
            seconds, but is not recommended. Homing helps the robot track the pipette's
            position.

        :returns: This instance.
        """
        alternate_drop_location: bool = False
        if location is None:
            trash_container = self.trash_container
            if self.api_version >= _DROP_TIP_LOCATION_ALTERNATING_ADDED_IN:
                alternate_drop_location = True
            if isinstance(trash_container, labware.Labware):
                well = trash_container.wells()[0]
            else:  # implicit drop tip in disposal location, not well
                with publisher.publish_context(
                    broker=self.broker,
                    command=cmds.drop_tip_in_disposal_location(
                        instrument=self, location=trash_container
                    ),
                ):
                    self._core.drop_tip_in_disposal_location(
                        trash_container,
                        home_after=home_after,
                    )
                self._last_tip_picked_up_from = None
                return self

        elif isinstance(location, labware.Well):
            well = location
            location = None

        elif isinstance(location, types.Location):
            _, maybe_well = location.labware.get_parent_labware_and_well()

            if maybe_well is None:
                raise TypeError(
                    "If a location is specified as a `types.Location`"
                    " (for instance, as the result of a call to `Well.top()`),"
                    " it must be a location relative to a well,"
                    " since that is where a tip is dropped."
                    f" However, the given location refers to {location.labware}"
                )

            well = maybe_well

        elif isinstance(location, (TrashBin, WasteChute)):
            with publisher.publish_context(
                broker=self.broker,
                command=cmds.drop_tip_in_disposal_location(
                    instrument=self, location=location
                ),
            ):
                self._core.drop_tip_in_disposal_location(
                    location, home_after=home_after
                )
            self._last_tip_picked_up_from = None
            return self

        else:
            raise TypeError(
                "If specified, location should be an instance of"
                " `types.Location` (e.g. the return value from `Well.top()`)"
                " or `Well` (e.g. `tiprack.wells()[0]`)."
                f" However, it is {location}"
            )

        with publisher.publish_context(
            broker=self.broker,
            command=cmds.drop_tip(instrument=self, location=well),
        ):
            self._core.drop_tip(
                location=location,
                well_core=well._core,
                home_after=home_after,
                alternate_drop_location=alternate_drop_location,
            )

        self._last_tip_picked_up_from = None
        return self

    @requires_version(2, 0)
    def home(self) -> InstrumentContext:
        """Home the robot.

        See :ref:`utility-homing`.

        :returns: This instance.
        """

        mount_name = self._core.get_mount().name.lower()

        with publisher.publish_context(
            broker=self.broker, command=cmds.home(mount_name)
        ):
            self._core.home()

        return self

    @requires_version(2, 0)
    def home_plunger(self) -> InstrumentContext:
        """Home the plunger associated with this mount.

        :returns: This instance.
        """
        self._core.home_plunger()
        return self

    # TODO (spp, 2024-03-08): verify if ok to & change source & dest types to AdvancedLiquidHandling
    @publisher.publish(command=cmds.distribute)
    @requires_version(2, 0)
    def distribute(
        self,
        volume: Union[float, Sequence[float]],
        source: labware.Well,
        dest: List[labware.Well],
        *args: Any,
        **kwargs: Any,
    ) -> InstrumentContext:
        """
        Move a volume of liquid from one source to multiple destinations.

        :param volume: The amount, in µL, to dispense into each destination well.
        :param source: A single well to aspirate liquid from.
        :param dest: A list of wells to dispense liquid into.
        :param kwargs: See :py:meth:`transfer` and the :ref:`complex_params` page.
            Some parameters behave differently than when transferring.

              - ``disposal_volume`` aspirates additional liquid to improve the accuracy
                of each dispense. Defaults to the minimum volume of the pipette. See
                :ref:`param-disposal-volume` for details.

              - ``mix_after`` is ignored.


        :returns: This instance.
        """
        _log.debug("Distributing {} from {} to {}".format(volume, source, dest))
        kwargs["mode"] = "distribute"
        kwargs["disposal_volume"] = kwargs.get("disposal_volume", self.min_volume)
        kwargs["mix_after"] = (0, 0)
        blowout_location = kwargs.get("blowout_location")
        instrument.validate_blowout_location(
            self.api_version, "distribute", blowout_location
        )

        return self.transfer(volume, source, dest, **kwargs)

    # TODO (spp, 2024-03-08): verify if ok to & change source & dest types to AdvancedLiquidHandling
    @publisher.publish(command=cmds.consolidate)
    @requires_version(2, 0)
    def consolidate(
        self,
        volume: Union[float, Sequence[float]],
        source: List[labware.Well],
        dest: labware.Well,
        *args: Any,
        **kwargs: Any,
    ) -> InstrumentContext:
        """
        Move liquid from multiple source wells to a single destination well.

        :param volume: The amount, in µL, to aspirate from each source well.
        :param source: A list of wells to aspirate liquid from.
        :param dest: A single well to dispense liquid into.
        :param kwargs: See :py:meth:`transfer` and the :ref:`complex_params` page.
                       Some parameters behave differently than when transferring.
                       ``disposal_volume`` and ``mix_before`` are ignored.
        :returns: This instance.
        """
        _log.debug("Consolidate {} from {} to {}".format(volume, source, dest))
        kwargs["mode"] = "consolidate"
        kwargs["mix_before"] = (0, 0)
        kwargs["disposal_volume"] = 0
        blowout_location = kwargs.get("blowout_location")
        instrument.validate_blowout_location(
            self.api_version, "consolidate", blowout_location
        )

        return self.transfer(volume, source, dest, **kwargs)

    @publisher.publish(command=cmds.transfer)
    @requires_version(2, 0)
    def transfer(  # noqa: C901
        self,
        volume: Union[float, Sequence[float]],
        source: AdvancedLiquidHandling,
        dest: AdvancedLiquidHandling,
        trash: bool = True,
        **kwargs: Any,
    ) -> InstrumentContext:
        # source: Union[Well, List[Well], List[List[Well]]],
        # dest: Union[Well, List[Well], List[List[Well]]],
        # TODO: Reach consensus on kwargs
        # TODO: Decide if to use a disposal_volume
        # TODO: Accordingly decide if remaining liquid should be blown out to
        # TODO: ..trash or the original well.
        # TODO: What should happen if the user passes a non-first-row well
        # TODO: ..as src/dest *while using multichannel pipette?
        """
        Move liquid from one well or group of wells to another.

        Transfer is a higher-level command, incorporating other
        :py:class:`InstrumentContext` commands, like :py:meth:`aspirate` and
        :py:meth:`dispense`. It makes writing a protocol easier at the cost of
        specificity. See :ref:`v2-complex-commands` for details on how transfer and
        other complex commands perform their component steps.

        :param volume: The amount, in µL, to aspirate from each source and dispense to
                       each destination. If ``volume`` is a list, each amount will be
                       used for the source and destination at the matching index. A list
                       item of ``0`` will skip the corresponding wells entirely. See
                       :ref:`complex-list-volumes` for details and examples.
        :param source: A single well or a list of wells to aspirate liquid from.
        :param dest: A single well or a list of wells to dispense liquid into.

        :Keyword Arguments: Transfer accepts a number of optional parameters that give
            you greater control over the exact steps it performs. See
            :ref:`complex_params` or the links under each argument's entry below for
            additional details and examples.

            * **new_tip** (*string*) --
              When to pick up and drop tips during the command. Defaults to ``"once"``.

                - ``"once"``: Use one tip for the entire command.
                - ``"always"``: Use a new tip for each set of aspirate and dispense steps.
                - ``"never"``: Do not pick up or drop tips at all.

              See :ref:`param-tip-handling` for details.

            * **trash** (*boolean*) --
              If ``True`` (default), the pipette will drop tips in its
              :py:meth:`~.InstrumentContext.trash_container`.
              If ``False``, the pipette will return tips to their tip rack.

              See :ref:`param-trash` for details.

            * **touch_tip** (*boolean*) --
              If ``True``, perform a :py:meth:`touch_tip` following each
              :py:meth:`aspirate` and :py:meth:`dispense`. Defaults to ``False``.

              See :ref:`param-touch-tip` for details.

            * **blow_out** (*boolean*) --
              If ``True``, a :py:meth:`blow_out` will occur following each
              :py:meth:`dispense`, but only if the pipette has no liquid left
              in it. If ``False`` (default), the pipette will not blow out liquid.

              See :ref:`param-blow-out` for details.

            * **blowout_location** (*string*) --
              Accepts one of three string values: ``"trash"``, ``"source well"``, or
              ``"destination well"``.

              If ``blow_out`` is ``False`` (its default), this parameter is ignored.

              If ``blow_out`` is ``True`` and this parameter is not set:

                - Blow out into the trash, if the pipette is empty or only contains the
                  disposal volume.

                - Blow out into the source well, if the pipette otherwise contains liquid.

            * **mix_before** (*tuple*) --
              Perform a :py:meth:`mix` before each :py:meth:`aspirate` during the
              transfer. The first value of the tuple is the number of repetitions, and
              the second value is the amount of liquid to mix in µL.

              See :ref:`param-mix-before` for details.

            * **mix_after** (*tuple*) --
              Perform a :py:meth:`mix` after each :py:meth:`dispense` during the
              transfer. The first value of the tuple is the number of repetitions, and
              the second value is the amount of liquid to mix in µL.

              See :ref:`param-mix-after` for details.

            * **disposal_volume** (*float*) --
              Transfer ignores the numeric value of this parameter. If set, the pipette
              will not aspirate additional liquid, but it will perform a very small blow
              out after each dispense.

              See :ref:`param-disposal-volume` for details.

        :returns: This instance.
        """
        _log.debug("Transfer {} from {} to {}".format(volume, source, dest))

        blowout_location = kwargs.get("blowout_location")
        instrument.validate_blowout_location(
            self.api_version, "transfer", blowout_location
        )

        kwargs["mode"] = kwargs.get("mode", "transfer")

        mix_strategy, mix_opts = mix_from_kwargs(kwargs)

        if trash:
            drop_tip = transfers.DropTipStrategy.TRASH
        else:
            drop_tip = transfers.DropTipStrategy.RETURN

        new_tip = kwargs.get("new_tip")
        if isinstance(new_tip, str):
            new_tip = types.TransferTipPolicy[new_tip.upper()]

        blow_out = kwargs.get("blow_out")
        blow_out_strategy = None
        active_channels = (
            self.active_channels
            if self._api_version >= _PARTIAL_NOZZLE_CONFIGURATION_ADDED_IN
            else self.channels
        )

        if blow_out and not blowout_location:
            if self.current_volume:
                blow_out_strategy = transfers.BlowOutStrategy.SOURCE
            else:
                blow_out_strategy = transfers.BlowOutStrategy.TRASH
        elif blow_out and blowout_location:
            if blowout_location == "source well":
                blow_out_strategy = transfers.BlowOutStrategy.SOURCE
            elif blowout_location == "destination well":
                blow_out_strategy = transfers.BlowOutStrategy.DEST
            elif blowout_location == "trash":
                blow_out_strategy = transfers.BlowOutStrategy.TRASH

        if new_tip != types.TransferTipPolicy.NEVER:
            tr, next_tip = labware.next_available_tip(
                self.starting_tip, self.tip_racks, active_channels
            )
            max_volume = min(next_tip.max_volume, self.max_volume)
        else:
            max_volume = self._core.get_working_volume()

        touch_tip = None
        if kwargs.get("touch_tip"):
            touch_tip = transfers.TouchTipStrategy.ALWAYS

        default_args = transfers.Transfer()

        disposal = kwargs.get("disposal_volume")
        if disposal is None:
            disposal = default_args.disposal_volume

        air_gap = kwargs.get("air_gap", default_args.air_gap)
        if air_gap < 0 or air_gap >= max_volume:
            raise ValueError(
                "air_gap must be between 0uL and the pipette's expected "
                f"working volume, {max_volume}uL"
            )

        transfer_args = transfers.Transfer(
            new_tip=new_tip or default_args.new_tip,
            air_gap=air_gap,
            carryover=kwargs.get("carryover") or default_args.carryover,
            gradient_function=(
                kwargs.get("gradient_function") or default_args.gradient_function
            ),
            disposal_volume=disposal,
            mix_strategy=mix_strategy,
            drop_tip_strategy=drop_tip,
            blow_out_strategy=blow_out_strategy or default_args.blow_out_strategy,
            touch_tip_strategy=(touch_tip or default_args.touch_tip_strategy),
        )
        transfer_options = transfers.TransferOptions(
            transfer=transfer_args, mix=mix_opts
        )
        plan = transfers.TransferPlan(
            volume,
            source,
            dest,
            self,
            max_volume,
            self.api_version,
            kwargs["mode"],
            transfer_options,
        )
        self._execute_transfer(plan)
        return self

    def _execute_transfer(self, plan: transfers.TransferPlan) -> None:
        for cmd in plan:
            getattr(self, cmd["method"])(*cmd["args"], **cmd["kwargs"])

    @requires_version(2, 0)
    def delay(self, *args: Any, **kwargs: Any) -> None:
        """
        .. deprecated:: 2.0
           Use :py:obj:`ProtocolContext.delay` instead.
           This method does nothing.
           It will be removed from a future version of the Python Protocol API.
        """
        if args or kwargs:
            # Former implementations of this method did not take any args, so users
            # would get a TypeError if they tried to call it like delay(minutes=10).
            # Without changing the ultimate behavior that such a call fails the
            # protocol, we can provide a more descriptive message as a courtesy.
            raise APIVersionError(
                "InstrumentContext.delay() is not supported in Python Protocol API v2."
                " Use ProtocolContext.delay() instead."
            )
        else:
            # Former implementations of this method, when called without any args,
            # called ProtocolContext.delay() with a duration of 0, which was
            # approximately a no-op.
            # Preserve that allowed way to call this method for the very remote chance
            # that a protocol out in the wild does it, for some reason.
            pass

    @requires_version(2, 0)
    def move_to(
        self,
        location: Union[types.Location, TrashBin, WasteChute],
        force_direct: bool = False,
        minimum_z_height: Optional[float] = None,
        speed: Optional[float] = None,
        publish: bool = True,
    ) -> InstrumentContext:
        """Move the instrument.

        See :ref:`move-to` for examples.

        :param location: The location to move to.
        :type location: :py:class:`~.types.Location`
        :param force_direct: If ``True``, move directly to the destination without arc
                             motion.

                             .. warning::
                                Forcing direct motion can cause the pipette to crash
                                into labware, modules, or other objects on the deck.

        :param minimum_z_height: An amount, measured in mm, to raise the mid-arc height.
                                 The mid-arc height can't be lowered.
        :param speed: The speed at which to move. By default,
                      :py:attr:`InstrumentContext.default_speed`. This controls the
                      straight linear speed of the motion. To limit individual axis
                      speeds, use :py:obj:`.ProtocolContext.max_speeds`.

        :param publish: Whether to list this function call in the run preview.
                        Default is ``True``.
        """
        with ExitStack() as contexts:
            if isinstance(location, (TrashBin, WasteChute)):
                if publish:
                    contexts.enter_context(
                        publisher.publish_context(
                            broker=self.broker,
                            command=cmds.move_to_disposal_location(
                                instrument=self, location=location
                            ),
                        )
                    )

                self._core.move_to(
                    location=location,
                    well_core=None,
                    force_direct=force_direct,
                    minimum_z_height=minimum_z_height,
                    speed=speed,
                )
            else:
                if publish:
                    contexts.enter_context(
                        publisher.publish_context(
                            broker=self.broker,
                            command=cmds.move_to(instrument=self, location=location),
                        )
                    )

                _, well = location.labware.get_parent_labware_and_well()

                self._core.move_to(
                    location=location,
                    well_core=well._core if well is not None else None,
                    force_direct=force_direct,
                    minimum_z_height=minimum_z_height,
                    speed=speed,
                )

        return self

    @property
    @requires_version(2, 0)
    def mount(self) -> str:
        """
        Return the name of the mount the pipette is attached to.

        The possible names are ``"left"`` and ``"right"``.
        """
        return self._core.get_mount().name.lower()

    @property
    @requires_version(2, 0)
    def speed(self) -> "PlungerSpeeds":
        """The speeds (in mm/s) configured for the pipette plunger.

        This is an object with attributes ``aspirate``, ``dispense``, and ``blow_out``
        holding the plunger speeds for the corresponding operation.

        .. note::
          Setting values of :py:attr:`flow_rate` will override the values in
          :py:attr:`speed`.

        .. versionchanged:: 2.14
            This property has been removed because it's fundamentally misaligned with
            the step-wise nature of a pipette's plunger speed configuration. Use
            :py:attr:`.flow_rate` instead.
        """
        if self._api_version >= ENGINE_CORE_API_VERSION:
            raise APIVersionError(
                "InstrumentContext.speed has been removed."
                " Use InstrumentContext.flow_rate, instead."
            )

        # TODO(mc, 2023-02-13): this assert should be enough for mypy
        # investigate if upgrading mypy allows the `cast` to be removed
        assert isinstance(self._core, LegacyInstrumentCore)
        return cast(LegacyInstrumentCore, self._core).get_speed()

    @property
    @requires_version(2, 0)
    def flow_rate(self) -> "FlowRates":
        """The speeds, in µL/s, configured for the pipette.

        See :ref:`new-plunger-flow-rates`.

        This is an object with attributes ``aspirate``, ``dispense``, and ``blow_out``
        holding the flow rate for the corresponding operation.

        .. note::
          Setting values of :py:attr:`speed`, which is deprecated, will override the
          values in :py:attr:`flow_rate`.

        """
        return self._core.get_flow_rate()

    @property
    @requires_version(2, 0)
    def type(self) -> str:
        """``'single'`` if this is a 1-channel pipette, or ``'multi'`` otherwise.

        See also :py:obj:`.channels`, which can distinguish between 8-channel and 96-channel
        pipettes.
        """
        if self.channels == 1:
            return "single"
        else:
            return "multi"

    @property
    @requires_version(2, 0)
    def tip_racks(self) -> List[labware.Labware]:
        """
        The tip racks that have been linked to this pipette.

        This is the property used to determine which tips to pick up next when calling
        :py:meth:`pick_up_tip` without arguments. See :ref:`basic-tip-pickup`.
        """
        return self._tip_racks

    @tip_racks.setter
    def tip_racks(self, racks: List[labware.Labware]) -> None:
        self._tip_racks = racks

    @property
    @requires_version(2, 0)
    def trash_container(self) -> Union[labware.Labware, TrashBin, WasteChute]:
        """The trash container associated with this pipette.

        This is the property used to determine where to drop tips and blow out liquids
        when calling :py:meth:`drop_tip` or :py:meth:`blow_out` without arguments.

        You can set this to a :py:obj:`Labware`, :py:class:`.TrashBin`, or :py:class:`.WasteChute`.

        The default value depends on the robot type and API version:

        - :py:obj:`ProtocolContext.fixed_trash`, if it exists.
        - Otherwise, the first item previously loaded with
          :py:obj:`ProtocolContext.load_trash_bin()` or
          :py:obj:`ProtocolContext.load_waste_chute()`.

        .. versionchanged:: 2.16
            Added support for ``TrashBin`` and ``WasteChute`` objects.
        """
        if self._user_specified_trash is None:
            disposal_locations = self._protocol_core.get_disposal_locations()
            if len(disposal_locations) == 0:
                raise NoTrashDefinedError(
                    "No trash container has been defined in this protocol."
                )
            return disposal_locations[0]
        return self._user_specified_trash

    @trash_container.setter
    def trash_container(
        self, trash: Union[labware.Labware, TrashBin, WasteChute]
    ) -> None:
        self._user_specified_trash = trash

    @property
    @requires_version(2, 0)
    def name(self) -> str:
        """
        The name string for the pipette (e.g., ``"p300_single"``).
        """
        return self._core.get_pipette_name()

    @property
    @requires_version(2, 0)
    def model(self) -> str:
        """
        The model string for the pipette (e.g., ``'p300_single_v1.3'``)
        """
        return self._core.get_model()

    @property
    @requires_version(2, 0)
    def min_volume(self) -> float:
        """
        The minimum volume, in µL, that the pipette can hold. This value may change
        based on the :ref:`volume mode <pipette-volume-modes>` that the pipette is
        currently configured for.
        """
        return self._core.get_min_volume()

    @property
    @requires_version(2, 0)
    def max_volume(self) -> float:
        """
        The maximum volume, in µL, that the pipette can hold.

        The maximum volume that you can actually aspirate might be lower than this,
        depending on what kind of tip is attached to this pipette. For example, a P300
        Single-Channel pipette always has a ``max_volume`` of 300 µL, but if it's using
        a 200 µL filter tip, its usable volume would be limited to 200 µL.
        """
        return self._core.get_max_volume()

    @property
    @requires_version(2, 0)
    def current_volume(self) -> float:
        """
        The current amount of liquid held in the pipette, measured in µL.
        """
        return self._core.get_current_volume()

    @property
    @requires_version(2, 7)
    def has_tip(self) -> bool:
        """Whether this instrument has a tip attached or not.

        The value of this property is determined logically by the API, not by detecting
        the physical presence of a tip. This is the case even on Flex, which has sensors
        to detect tip attachment.
        """
        return self._core.has_tip()

    @property
    def _has_tip(self) -> bool:
        """
        Internal function used to check whether this instrument has a
        tip attached or not.
        """
        return self._core.has_tip()

    @property
    @requires_version(2, 0)
    def hw_pipette(self) -> PipetteDict:
        """View the information returned by the hardware API directly.

        :raises: :py:class:`.types.PipetteNotAttachedError` if the pipette is
                 no longer attached (should not happen).
        """
        return self._core.get_hardware_state()

    @property
    @requires_version(2, 0)
    def channels(self) -> int:
        """The number of channels on the pipette.

        Possible values are 1, 8, or 96.

        See also :py:obj:`.type`.
        """
        return self._core.get_channels()

    @property
    @requires_version(2, 16)
    def active_channels(self) -> int:
        """The number of channels the pipette will use to pick up tips.

        By default, all channels on the pipette. Use :py:meth:`.configure_nozzle_layout`
        to set the pipette to use fewer channels.
        """
        return self._core.get_active_channels()

    @property
    @requires_version(2, 2)
    def return_height(self) -> float:
        """The height to return a tip to its tip rack.

        :returns: A scaling factor to apply to the tip length.
                  During :py:meth:`.drop_tip`, this factor is multiplied by the tip
                  length to get the distance from the top of the well to drop the tip.
        """
        return self._core.get_return_height()

    @property
    @requires_version(2, 0)
    def well_bottom_clearance(self) -> "Clearances":
        """The distance above the bottom of a well to aspirate or dispense.

        This is an object with attributes ``aspirate`` and ``dispense``, describing the
        default height of the corresponding operation. The default is 1.0 mm for both
        aspirate and dispense.

        When :py:meth:`aspirate` or :py:meth:`dispense` is given a :py:class:`.Well`
        rather than a full :py:class:`.Location`, the robot will move this distance
        above the bottom of the well to aspirate or dispense.

        To change, set the corresponding attribute::

            pipette.well_bottom_clearance.aspirate = 2

        """
        return self._well_bottom_clearances

    def _get_last_location_by_api_version(self) -> Optional[types.Location]:
        """Get the last location accessed by this pipette, if any.

        In pre-engine Protocol API versions, this call omits the pipette mount.
        This is to preserve pre-existing, potentially buggy behavior.
        """
        if self._api_version >= ENGINE_CORE_API_VERSION:
            return self._protocol_core.get_last_location(mount=self._core.get_mount())
        else:
            return self._protocol_core.get_last_location()

    def __repr__(self) -> str:
        return "<{}: {} in {}>".format(
            self.__class__.__name__,
            self._core.get_model(),
            self._core.get_mount().name,
        )

    def __str__(self) -> str:
        return "{} on {} mount".format(self._core.get_display_name(), self.mount)

    @requires_version(2, 15)
    def configure_for_volume(self, volume: float) -> None:
        """Configure a pipette to handle a specific volume of liquid, measured in µL.
        The pipette enters a volume mode depending on the volume provided. Changing
        pipette modes alters properties of the instance of
        :py:class:`.InstrumentContext`, such as default flow rate, minimum volume, and
        maximum volume. The pipette remains in the mode set by this function until it is
        called again.

        The Flex 1-Channel 50 µL and Flex 8-Channel 50 µL pipettes must operate in a
        low-volume mode to accurately dispense very small volumes of liquid. Low-volume
        mode can only be set by calling ``configure_for_volume()``. See
        :ref:`pipette-volume-modes`.

        .. note ::

            Changing a pipette's mode will reset its :ref:`flow rates
            <new-plunger-flow-rates>`.

        This function will raise an error if called when the pipette's tip contains
        liquid. It won't raise an error if a tip is not attached, but changing modes may
        affect which tips the pipette can subsequently pick up without raising an error.

        This function will also raise an error if ``volume`` is outside of the
        :ref:`minimum and maximum capacities <new-pipette-models>` of the pipette (e.g.,
        setting ``volume=1`` for a Flex 1000 µL pipette).

        :param volume: The volume, in µL, that the pipette will prepare to handle.
        :type volume: float
        """
        if self._core.get_current_volume():
            raise CommandPreconditionViolated(
                message=f"Cannot switch modes of {str(self)} while it contains liquid"
            )
        if volume < 0:
            raise CommandParameterLimitViolated(
                command_name="configure_for_volume",
                parameter_name="volume",
                limit_statement="must be greater than 0",
                actual_value=str(volume),
            )
        last_location = self._get_last_location_by_api_version()
        if last_location and isinstance(last_location.labware, labware.Well):
            self.move_to(last_location.labware.top())
        self._core.configure_for_volume(volume)

    @requires_version(2, 16)
    def prepare_to_aspirate(self) -> None:
        """Prepare a pipette for aspiration.

        Before a pipette can aspirate into an empty tip, the plunger must be in its
        bottom position. After dropping a tip or blowing out, the plunger will be in a
        different position. This function moves the plunger to the bottom position,
        regardless of its current position, to make sure that the pipette is ready to
        aspirate.

        You rarely need to call this function. The API automatically prepares the
        pipette for aspiration as part of other commands:

            - After picking up a tip with :py:meth:`.pick_up_tip`.
            - When calling :py:meth:`.aspirate`, if the pipette isn't already prepared.
              If the pipette is in a well, it will move out of the well, move the plunger,
              and then move back.

        Use ``prepare_to_aspirate`` when you need to control exactly when the plunger
        motion will happen. A common use case is a pre-wetting routine, which requires
        preparing for aspiration, moving into a well, and then aspirating *without
        leaving the well*::

             pipette.move_to(well.bottom(z=2))
             pipette.delay(5)
             pipette.mix(10, 10)
             pipette.move_to(well.top(z=5))
             pipette.blow_out()
             pipette.prepare_to_aspirate()
             pipette.move_to(well.bottom(z=2))
             pipette.delay(5)
             pipette.aspirate(10, well.bottom(z=2))

        The call to ``prepare_to_aspirate()`` means that the plunger will be in the
        bottom position before the call to ``aspirate()``. Since it doesn't need to
        prepare again, it will not move up out of the well to move the plunger. It will
        aspirate in place.
        """
        if self._core.get_current_volume():
            raise CommandPreconditionViolated(
                message=f"Cannot prepare {str(self)} for aspirate while it contains liquid."
            )
        self._core.prepare_to_aspirate()

    @requires_version(2, 16)
    def configure_nozzle_layout(
        self,
        style: NozzleLayout,
        start: Optional[str] = None,
        front_right: Optional[str] = None,
        tip_racks: Optional[List[labware.Labware]] = None,
    ) -> None:
        """Configure how many tips the 96-channel pipette will pick up.

        Changing the nozzle layout will affect gantry movement for all subsequent
        pipetting actions that the pipette performs. It also alters the pipette's
        behavior for picking up tips. The pipette will continue to use the specified
        layout until this function is called again.

        .. note::
            When picking up fewer than 96 tips at once, the tip rack *must not* be
            placed in a tip rack adapter in the deck. If you try to pick up fewer than 96
            tips from a tip rack that is in an adapter, the API will raise an error.

        :param style: The shape of the nozzle layout.

            - ``COLUMN`` sets the pipette to use 8 nozzles, aligned from front to back
              with respect to the deck. This corresponds to a column of wells on labware.
            - ``ALL`` resets the pipette to use all of its nozzles. Calling
              ``configure_nozzle_layout`` with no arguments also resets the pipette.

        :type style: ``NozzleLayout`` or ``None``
        :param start: The nozzle at the back left of the layout, which the robot uses
            to determine how it will move to different locations on the deck. The string
            should be of the same format used when identifying wells by name.
            Required unless setting ``style=ALL``.

            .. note::
                When using the ``COLUMN`` layout, the only fully supported value is
                ``start="A12"``. You can use ``start="A1"``, but this will disable tip
                tracking and you will have to specify the ``location`` every time you
                call :py:meth:`.pick_up_tip`, such that the pipette picks up columns of
                tips *from right to left* on the tip rack.

        :type start: str or ``None``
        :param tip_racks: Behaves the same as setting the ``tip_racks`` parameter of
            :py:meth:`.load_instrument`. If not specified, the new configuration resets
            :py:obj:`.InstrumentContext.tip_racks` and you must specify the location
            every time you call :py:meth:`~.InstrumentContext.pick_up_tip`.
        :type tip_racks: List[:py:class:`.Labware`]
        """
        #       TODO: add the following back into the docstring when QUADRANT is supported
        #
        #       :param front_right: The nozzle at the front left of the layout. Only used for
        #           NozzleLayout.QUADRANT configurations.
        #       :type front_right: str or ``None``
        if style != NozzleLayout.ALL:
            if start is None:
                raise ValueError(
                    f"Cannot configure a nozzle layout of style {style.value} without a starting nozzle."
                )
            if start not in types.ALLOWED_PRIMARY_NOZZLES:
                raise ValueError(
                    f"Starting nozzle specified is not one of {types.ALLOWED_PRIMARY_NOZZLES}"
                )
        if style == NozzleLayout.QUADRANT:
            if front_right is None:
                raise ValueError(
                    "Cannot configure a QUADRANT layout without a front right nozzle."
                )
        self._core.configure_nozzle_layout(
            style,
            primary_nozzle=start,
            front_right_nozzle=front_right,
        )
        # TODO (spp, 2023-12-05): verify that tipracks are on adapters for only full 96 channel config
        self._tip_racks = tip_racks or []



from __future__ import annotations

""" Classes and functions for gripper state tracking
"""
import logging
from typing import Any, Optional, Set, Dict, Tuple, Final

from opentrons.types import Point
from opentrons.config import gripper_config
from opentrons.hardware_control.types import (
    GripperProbe,
    CriticalPoint,
    GripperJawState,
)
from opentrons.hardware_control.errors import (
    InvalidCriticalPoint,
)
from .instrument_calibration import (
    GripperCalibrationOffset,
    load_gripper_calibration_offset,
    save_gripper_calibration_offset,
)
from ..instrument_abc import AbstractInstrument
from opentrons.hardware_control.dev_types import AttachedGripper, GripperDict
from opentrons_shared_data.errors.exceptions import (
    CommandPreconditionViolated,
    MotionFailedError,
)

from opentrons_shared_data.gripper import (
    GripperDefinition,
    GripForceProfile,
    Geometry,
)

RECONFIG_KEYS = {"quirks"}

MAX_ACCEPTABLE_JAW_DISPLACEMENT: Final = 20

mod_log = logging.getLogger(__name__)


class Gripper(AbstractInstrument[GripperDefinition]):
    """A class to gather and track gripper state and configs.

    This class should not touch hardware or call back out to the hardware
    control API. Its only purpose is to gather state.
    """

    def __init__(
        self,
        config: GripperDefinition,
        gripper_cal_offset: GripperCalibrationOffset,
        gripper_id: str,
    ) -> None:
        self._config = config
        self._model = config.model

        self._geometry = self._config.geometry
        base_offset = Point(*self._geometry.base_offset_from_mount)
        self._jaw_center_offset = (
            Point(*self._geometry.jaw_center_offset_from_base) + base_offset
        )
        #: the distance between the gripper mount and the jaw center at home
        self._front_calibration_pin_offset = (
            Point(*self._geometry.pin_one_offset_from_base) + base_offset
        )
        #: the distance between the gripper mount and the front calibration pin
        #: at home
        self._rear_calibration_pin_offset = (
            Point(*self._geometry.pin_two_offset_from_base) + base_offset
        )
        #: the distance between the gripper mount and the rear calibration pin
        #: at home
        self._calibration_offset = gripper_cal_offset
        #: The output value of calibration - the additional vector added into
        #: the critical point geometry based on gripper mount calibration
        self._gripper_id = gripper_id
        self._state = GripperJawState.UNHOMED
        self._current_jaw_displacement = 0.0
        self._attached_probe: Optional[GripperProbe] = None
        self._log = mod_log.getChild(self._gripper_id)
        self._log.info(
            f"loaded: {self._model}, gripper offset: {self._calibration_offset}"
        )
        self._jaw_max_offset: Optional[float] = None

    @property
    def grip_force_profile(self) -> GripForceProfile:
        return self._config.grip_force_profile

    @property
    def geometry(self) -> Geometry:
        return self._geometry

    @property
    def attached_probe(self) -> Optional[GripperProbe]:
        return self._attached_probe

    def add_probe(self, probe: GripperProbe) -> None:
        """This is used for finding the critical point during calibration."""
        assert not self.attached_probe
        self._attached_probe = probe

    def remove_probe(self) -> None:
        assert self.attached_probe
        self._attached_probe = None

    @property
    def max_allowed_grip_error(self) -> float:
        return self._geometry.max_allowed_grip_error

    @property
    def max_jaw_width(self) -> float:
        return self._config.geometry.jaw_width["max"] + (self._jaw_max_offset or 0)

    @property
    def min_jaw_width(self) -> float:
        return self._config.geometry.jaw_width["min"]

    @property
    def jaw_width(self) -> float:
        return self.max_jaw_width - (self.current_jaw_displacement * 2.0)

    @property
    def current_jaw_displacement(self) -> float:
        """The distance one side of the jaw has traveled from home."""
        return self._current_jaw_displacement

    @current_jaw_displacement.setter
    def current_jaw_displacement(self, mm: float) -> None:
        max_mm = self.max_jaw_displacement() + 2.0
        if mm > max_mm:
            self._log.warning(
                f"jaw displacement {round(mm, 1)} mm exceeds max expected value: "
                f"{max_mm} mm, setting value to max value instead."
            )
        self._current_jaw_displacement = min(mm, max_mm)

    @property
    def default_grip_force(self) -> float:
        return self.grip_force_profile.default_grip_force

    @property
    def default_idle_force(self) -> float:
        return self.grip_force_profile.default_idle_force

    @property
    def default_home_force(self) -> float:
        return self.grip_force_profile.default_home_force

    def max_jaw_displacement(self) -> float:
        geometry = self._config.geometry
        return (self.max_jaw_width - geometry.jaw_width["min"]) / 2

    @property
    def state(self) -> GripperJawState:
        return self._state

    @state.setter
    def state(self, s: GripperJawState) -> None:
        self._state = s

    @property
    def config(self) -> GripperDefinition:
        return self._config

    def update_config_item(self, elements: Dict[str, Any]) -> None:
        raise NotImplementedError("Update config is not supported at this time.")

    @property
    def model(self) -> str:
        return repr(self._model)

    @property
    def gripper_id(self) -> str:
        return self._gripper_id

    def reload_configurations(self) -> None:
        return None

    def update_jaw_open_position_from_closed_position(
        self, jaw_at_closed: float
    ) -> None:
        """Update the estimation of the jaw position at open based on reading it at closed.

        This is necessary because the gripper jaw has a well-defined positional hard stop
        when fully closed and empty but _not_ when open. The open position can vary unit to
        unit. You can calibrate this out by reading the position of the encoder when the jaw
        is closed, and then altering the logical open position so that it is whatever it needs
        to be for the logical closed position to be the same as the config.
        """
        jaw_min = self._config.geometry.jaw_width["min"]
        jaw_nominal_max = self._config.geometry.jaw_width["max"]
        if (
            abs((jaw_at_closed * 2) - (jaw_nominal_max - jaw_min))
            > MAX_ACCEPTABLE_JAW_DISPLACEMENT
        ):
            raise MotionFailedError(
                message="Gripper jaw calibration out of bounds",
                detail={
                    "type": "gripper-jaw-calibration-out-of-bounds",
                    "actual-displacement": str(jaw_at_closed * 2),
                    "nominal-displacement": str(jaw_nominal_max - jaw_min),
                },
            )

        self._jaw_max_offset = jaw_min - (jaw_nominal_max - (jaw_at_closed * 2))
        self._log.info(
            f"Gripper max jaw offset is now {self._jaw_max_offset} from input position {jaw_at_closed}"
        )

    @property
    def has_jaw_width_calibration(self) -> bool:
        return self._jaw_max_offset is not None

    def reset_offset(self, to_default: bool) -> None:
        """Tempoarily reset the gripper offsets to default values."""
        if to_default:
            self._calibration_offset = load_gripper_calibration_offset(gripper_id=None)
        else:
            self._calibration_offset = load_gripper_calibration_offset(
                gripper_id=self._gripper_id
            )

    def save_offset(self, delta: Point) -> GripperCalibrationOffset:
        """Save a new gripper offset."""
        save_gripper_calibration_offset(self.gripper_id, delta)
        self._calibration_offset = load_gripper_calibration_offset(self.gripper_id)
        return self._calibration_offset

    def check_calibration_pin_location_is_accurate(self) -> None:
        if not self.attached_probe:
            raise CommandPreconditionViolated(
                "Cannot calibrate gripper without attaching a calibration probe",
                detail={
                    "probe": str(self._attached_probe),
                    "jaw_state": str(self.state),
                },
            )
        if self.state != GripperJawState.GRIPPING:
            raise CommandPreconditionViolated(
                "Cannot calibrate gripper if jaw is not in gripping state",
                detail={
                    "probe": str(self._attached_probe),
                    "jaw_state": str(self.state),
                },
            )

    def critical_point(self, cp_override: Optional[CriticalPoint] = None) -> Point:
        """
        The vector from the gripper mount to the critical point, which is selectable
        between the center of the gripper engagement volume and the calibration pins.
        """
        if cp_override in [CriticalPoint.NOZZLE, CriticalPoint.TIP]:
            raise InvalidCriticalPoint(cp_override.name, "gripper")

        if not self._attached_probe:
            cp = cp_override or CriticalPoint.GRIPPER_JAW_CENTER
        else:
            if self._attached_probe is GripperProbe.REAR:
                cp = cp_override or CriticalPoint.GRIPPER_REAR_CALIBRATION_PIN
            else:
                cp = cp_override or CriticalPoint.GRIPPER_FRONT_CALIBRATION_PIN

        if cp in [CriticalPoint.GRIPPER_JAW_CENTER, CriticalPoint.XY_CENTER]:
            return self._jaw_center_offset + Point(*self._calibration_offset.offset)
        elif cp == CriticalPoint.GRIPPER_FRONT_CALIBRATION_PIN:
            self.check_calibration_pin_location_is_accurate()
            return (
                self._front_calibration_pin_offset
                + Point(*self._calibration_offset.offset)
                + Point(y=self.current_jaw_displacement)
                - Point(y=(self._jaw_max_offset or 0))
            )
        elif cp == CriticalPoint.GRIPPER_REAR_CALIBRATION_PIN:
            self.check_calibration_pin_location_is_accurate()
            return (
                self._rear_calibration_pin_offset
                + Point(*self._calibration_offset.offset)
                - Point(y=self.current_jaw_displacement)
                + Point(y=(self._jaw_max_offset or 0))
            )
        else:
            raise InvalidCriticalPoint(cp.name, "gripper")

    def duty_cycle_by_force(self, newton: float) -> float:
        return gripper_config.duty_cycle_by_force(newton, self.grip_force_profile)

    def __str__(self) -> str:
        return f"{self._config.display_name}"

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self._config.display_name} {id(self)}"

    def as_dict(self) -> GripperDict:
        d: GripperDict = {
            "model": self._config.model,
            "gripper_id": self._gripper_id,
            "display_name": self._config.display_name,
            "state": self._state,
            "calibration_offset": self._calibration_offset,
        }
        return d


def _reload_gripper(
    new_config: GripperDefinition,
    attached_instr: Gripper,
    cal_offset: GripperCalibrationOffset,
) -> Tuple[Gripper, bool]:
    # Once we have determined that the new and attached grippers
    # are similar enough that we might skip, see if the configs
    # match closely enough.
    # Returns a gripper object
    if new_config == attached_instr.config:
        # Same config, good enough
        return attached_instr, True
    else:
        newdict = new_config.dict()
        olddict = attached_instr.config.dict()
        changed: Set[str] = set()
        for k in newdict.keys():
            if newdict[k] != olddict[k]:
                changed.add(k)
        if changed.intersection(RECONFIG_KEYS):
            # Something has changed that requires reconfig
            return (
                Gripper(
                    new_config,
                    cal_offset,
                    attached_instr._gripper_id,
                ),
                False,
            )
        else:
            # update just the cal offset and update info
            attached_instr._calibration_offset = cal_offset
            return attached_instr, True


def compare_gripper_config_and_check_skip(
    freshly_detected: AttachedGripper,
    attached: Optional[Gripper],
    cal_offset: GripperCalibrationOffset,
) -> Tuple[Optional[Gripper], bool]:
    """
    Given the gripper config for an attached gripper (if any) freshly read
    from disk, and any attached instruments,

    - Compare the new and configured gripper configs
    - Load the new configs if they differ
    - Return a bool indicating whether hardware reconfiguration may be
      skipped
    """
    config = freshly_detected.get("config")
    serial = freshly_detected.get("id") or ""

    if not config and not attached:
        # nothing attached now, nothing used to be attached, nothing
        # to reconfigure
        return attached, True

    if config and attached:
        # something was attached and something is attached. are they
        # the same? we can tell by comparing serials
        if serial == attached.gripper_id:
            # similar enough to check
            return _reload_gripper(config, attached, cal_offset)

    if config:
        return Gripper(config, cal_offset, serial), False
    else:
        return None, False



"""Hardware Modules' substates."""

from typing import Union
from .magnetic_module_substate import MagneticModuleSubState, MagneticModuleId
from .heater_shaker_module_substate import (
    HeaterShakerModuleSubState,
    HeaterShakerModuleId,
)
from .temperature_module_substate import TemperatureModuleSubState, TemperatureModuleId
from .thermocycler_module_substate import (
    ThermocyclerModuleSubState,
    ThermocyclerModuleId,
)
from .magnetic_block_substate import MagneticBlockSubState, MagneticBlockId


ModuleSubStateType = Union[
    HeaterShakerModuleSubState,
    MagneticModuleSubState,
    TemperatureModuleSubState,
    ThermocyclerModuleSubState,
    MagneticBlockSubState,
]

__all__ = [
    "MagneticModuleSubState",
    "MagneticModuleId",
    "HeaterShakerModuleSubState",
    "HeaterShakerModuleId",
    "TemperatureModuleSubState",
    "TemperatureModuleId",
    "ThermocyclerModuleSubState",
    "ThermocyclerModuleId",
    "MagneticBlockSubState",
    "MagneticBlockId",
    # Union of all module substates
    "ModuleSubStateType",
]



from typing import List
from typing_extensions import Protocol

from opentrons.hardware_control.modules.types import ModuleAtPort


class USBDriverInterface(Protocol):
    def match_virtual_ports(
        self,
        virtual_port: List[ModuleAtPort],
    ) -> List[ModuleAtPort]:
        ...



from typing import Optional, Dict, Union

from opentrons.drivers.smoothie_drivers.constants import (
    HOMED_POSITION,
    Y_BOUND_OVERRIDE,
    DISABLE_AXES,
    AXES,
)
from opentrons.drivers.types import MoveSplits


class SimulatingDriver:
    def __init__(self) -> None:
        self._steps_per_mm: Dict[str, float] = {}
        self._homed_flags: Dict[str, bool] = {}

    @property
    def homed_flags(self) -> Dict[str, bool]:
        return self._homed_flags

    async def home(self, axis: str = AXES, disabled: str = DISABLE_AXES) -> None:
        for ax in axis:
            self._homed_flags[ax] = True

    async def _smoothie_reset(self) -> None:
        self._homed_flags.clear()

    async def read_pipette_id(self, mount: str) -> Optional[str]:
        pass

    async def read_pipette_model(self, mount: str) -> Optional[str]:
        pass

    async def write_pipette_id(self, mount: str, data_string: str) -> None:
        pass

    async def write_pipette_model(self, mount: str, data_string: str) -> None:
        pass

    # def _send_command(self, command, timeout=None):
    #     pass

    def turn_on_blue_button_light(self) -> None:
        pass

    def turn_on_red_button_light(self) -> None:
        pass

    async def update_pipette_config(
        self, axis: str, data: Dict[str, float]
    ) -> Dict[str, Dict[str, float]]:
        """
        Updates the following configs for a given pipette mount based on
        the detected pipette type:
        - homing positions M365.0
        - Max Travel M365.1
        - endstop debounce M365.2 (NOT for zprobe debounce)
        - retract from endstop distance M365.3
        """
        return {}

    @property
    def current(self) -> Dict[str, float]:
        return {}

    @property
    def speed(self) -> None:
        pass

    @property
    def steps_per_mm(self) -> Dict[str, float]:
        return self._steps_per_mm

    # @steps_per_mm.setter
    # def steps_per_mm(self, axis, mm):
    #     # Keep track of any updates to the steps per mm per axis
    #     self._steps_per_mm[axis] = mm

    async def update_steps_per_mm(self, data: Union[Dict[str, float], str]) -> None:
        pass

    def configure_splits_for(self, config: MoveSplits) -> None:
        pass

    def set_dwelling_current(self, settings: Dict[str, float]) -> None:
        pass

    def set_acceleration(self, settings: Dict[str, float]) -> None:
        pass

    @property
    def homed_position(self) -> Dict[str, float]:
        return HOMED_POSITION.copy()

    @property
    def axis_bounds(self) -> Dict[str, float]:
        position = HOMED_POSITION.copy()
        position["Y"] = Y_BOUND_OVERRIDE
        return position



from abc import ABC, abstractmethod
from typing import Any, Optional, Generic, TypeVar, Dict

from opentrons.types import Point
from opentrons.hardware_control.types import CriticalPoint


InstrumentConfig = TypeVar("InstrumentConfig")


class AbstractInstrument(ABC, Generic[InstrumentConfig]):
    """Defines the common methods of an instrument."""

    @property
    @abstractmethod
    def model(self) -> str:
        """Return model of the instrument."""
        ...

    @property
    @abstractmethod
    def config(self) -> InstrumentConfig:
        """Instrument config in dataclass format."""
        ...

    @abstractmethod
    def reload_configurations(self) -> None:
        """Reset the instrument to default configurations."""
        ...

    @abstractmethod
    def update_config_item(self, elements: Dict[str, Any]) -> None:
        """Update instrument config item."""
        ...

    @abstractmethod
    def critical_point(self, cp_override: Optional[CriticalPoint] = None) -> Point:
        """Computate critical point of an instrument."""
        ...
